# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import sys
import threading
import traceback
from typing import Dict, List

from ... import oscar as mo
from ...oscar.backends.allocate_strategy import ProcessIndex


class ProcessInfoManagerActor(mo.StatelessActor):
    _process_refs: List[mo.ActorRef]

    def __init__(self):
        self._process_refs = []
        self._pool_configs = []

    async def __post_create__(self):
        index = 0
        while True:
            try:
                ref = await mo.create_actor(
                    ProcessInfoActor,
                    process_index=index,
                    uid=ProcessInfoActor.gen_uid(index),
                    address=self.address,
                    allocate_strategy=ProcessIndex(index),
                )
            except IndexError:
                break

            index += 1
            self._process_refs.append(ref)

        self._pool_configs = await asyncio.gather(
            *[ref.get_pool_config() for ref in self._process_refs]
        )

    async def get_pool_configs(self) -> List[Dict]:
        return self._pool_configs

    async def get_thread_stacks(self) -> List[Dict[int, List[str]]]:
        stack_tasks = [
            asyncio.create_task(ref.get_thread_stacks()) for ref in self._process_refs
        ]
        await asyncio.wait(stack_tasks, return_when=asyncio.ALL_COMPLETED)

        results = []
        for fut in stack_tasks:
            try:
                results.append(fut.result())
            except (mo.ActorNotExist, mo.ServerClosed):
                results.append(None)
        return results


class ProcessInfoActor(mo.StatelessActor):
    def __init__(self, process_index: int = 0):
        self._process_index = process_index
        self._pool_config = None

    async def __post_create__(self):
        self._pool_config = await mo.get_pool_config(self.address)

    @classmethod
    def gen_uid(cls, process_index: int):
        return f"process_info_{process_index}"

    def get_pool_config(self) -> dict:
        idx = self._pool_config.get_process_index(self.address)
        return self._pool_config.get_pool_config(idx)

    @classmethod
    def get_thread_stacks(cls) -> Dict[str, List[str]]:
        frames = sys._current_frames()
        stacks = dict()
        for th in threading.enumerate():
            tid = getattr(th, "native_id", th.ident)
            stack_key = f"{tid}:{th.name}"
            stacks[stack_key] = traceback.format_stack(frames[th.ident])
        return stacks



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import AnyField, Int32Field, BoolField, StringField
from ...tensor.utils import filter_inputs
from ..core import DATAFRAME_TYPE, SERIES_TYPE
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import build_df, build_series, validate_axis


class DataFrameWhere(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.WHERE

    _input = AnyField("input")
    _cond = AnyField("cond")
    _other = AnyField("other")
    _axis = Int32Field("axis")
    _level = AnyField("level")
    _errors = StringField("errors")
    _try_cast = BoolField("try_cast")
    _replace_true = BoolField("replace_true")

    def __init__(
        self,
        input=None,
        cond=None,
        other=None,  # pylint: disable=redefined-builtin
        axis=None,
        level=None,
        errors=None,
        try_cast=None,
        replace_true=None,
        **kw
    ):
        super().__init__(
            _input=input,
            _cond=cond,
            _other=other,
            _axis=axis,
            _level=level,
            _errors=errors,
            _try_cast=try_cast,
            _replace_true=replace_true,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def cond(self):
        return self._cond

    @property
    def other(self):
        return self._other

    @property
    def axis(self):
        return self._axis

    @property
    def level(self):
        return self._level

    @property
    def errors(self):
        return self._errors

    @property
    def try_cast(self):
        return self._try_cast

    @property
    def replace_true(self):
        return self._replace_true

    def __call__(self, df_or_series):
        def _check_input_index(obj, axis=None):
            axis = axis if axis is not None else self.axis
            if isinstance(obj, DATAFRAME_TYPE) and (
                df_or_series.columns_value.key != obj.columns_value.key
                or df_or_series.index_value.key != obj.index_value.key
            ):
                raise NotImplementedError("Aligning different indices not supported")
            elif (
                isinstance(obj, SERIES_TYPE)
                and df_or_series.axes[axis].index_value.key != obj.index_value.key
            ):
                raise NotImplementedError("Aligning different indices not supported")

        _check_input_index(self.cond, axis=0)
        _check_input_index(self.other)

        if isinstance(df_or_series, DATAFRAME_TYPE):
            mock_obj = build_df(df_or_series)
        else:
            mock_obj = build_series(df_or_series)

        if isinstance(self.other, (pd.DataFrame, DATAFRAME_TYPE)):
            mock_other = build_df(self.other)
        elif isinstance(self.other, (pd.Series, SERIES_TYPE)):
            mock_other = build_series(self.other)
        else:
            mock_other = self.other

        result_df = mock_obj.where(
            np.zeros(mock_obj.shape).astype(bool),
            other=mock_other,
            axis=self.axis,
            level=self.level,
            errors=self.errors,
            try_cast=self.try_cast,
        )

        inputs = filter_inputs([df_or_series, self.cond, self.other])
        if isinstance(df_or_series, DATAFRAME_TYPE):
            return self.new_dataframe(
                inputs,
                shape=df_or_series.shape,
                dtypes=result_df.dtypes,
                index_value=df_or_series.index_value,
                columns_value=df_or_series.columns_value,
            )
        else:
            return self.new_series(
                inputs,
                shape=df_or_series.shape,
                name=df_or_series.name,
                dtype=result_df.dtype,
                index_value=df_or_series.index_value,
            )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        if isinstance(self._cond, ENTITY_TYPE):
            self._cond = next(inputs_iter)
        if isinstance(self._other, ENTITY_TYPE):
            self._other = next(inputs_iter)

    @classmethod
    def tile(cls, op: "DataFrameWhere"):
        def rechunk_input(inp, axis=None):
            axis = axis if axis is not None else op.axis
            if isinstance(inp, DATAFRAME_TYPE):
                inp = yield from recursive_tile(inp.rechunk(op.input.nsplits))
            elif isinstance(inp, SERIES_TYPE):
                inp = yield from recursive_tile(
                    inp.rechunk({0: op.input.nsplits[axis]})
                )
            return inp

        def get_tiled_chunk(obj, index, axis=None):
            if isinstance(obj, DATAFRAME_TYPE):
                return obj.cix[index[0], index[1]]
            elif isinstance(obj, SERIES_TYPE):
                axis = axis if axis is not None else op.axis
                return obj.cix[
                    index[axis],
                ]
            else:
                return obj

        # TODO support axis alignment for three objects
        cond = yield from rechunk_input(op.cond, axis=0)
        other = yield from rechunk_input(op.other)

        chunks = []
        for c in op.input.chunks:
            cond_chunk = get_tiled_chunk(cond, c.index, axis=0)
            other_chunk = get_tiled_chunk(other, c.index)

            new_op = op.copy().reset_key()
            new_op._cond = cond_chunk
            new_op._other = other_chunk

            inputs = filter_inputs([c, cond_chunk, other_chunk])
            chunks.append(new_op.new_chunk(inputs, **c.params))

        new_op = op.copy().reset_key()
        return new_op.new_tileables(
            op.inputs, chunks=chunks, nsplits=op.input.nsplits, **op.input.params
        )

    @classmethod
    def execute(cls, ctx, op: "DataFrameWhere"):
        out_obj = op.outputs[0]

        input_data = ctx[op.input.key]
        cond = op.cond
        if isinstance(cond, ENTITY_TYPE):
            cond = ctx[cond.key]

        other = op.other
        if isinstance(other, ENTITY_TYPE):
            other = ctx[other.key]

        if op.replace_true:
            ctx[out_obj.key] = input_data.mask(
                cond,
                other,
                axis=op.axis,
                level=op.level,
                errors=op.errors,
                try_cast=op.try_cast,
            )
        else:
            ctx[out_obj.key] = input_data.where(
                cond,
                other,
                axis=op.axis,
                level=op.level,
                errors=op.errors,
                try_cast=op.try_cast,
            )


_doc_template = """
Replace values where the condition is {replace_true}.

Parameters
----------
cond : bool Series/DataFrame, array-like, or callable
    Where `cond` is False, keep the original value. Where
    True, replace with corresponding value from `other`.
    If `cond` is callable, it is computed on the Series/DataFrame and
    should return boolean Series/DataFrame or array. The callable must
    not change input Series/DataFrame (though pandas doesn't check it).
other : scalar, Series/DataFrame, or callable
    Entries where `cond` is True are replaced with
    corresponding value from `other`.
    If other is callable, it is computed on the Series/DataFrame and
    should return scalar or Series/DataFrame. The callable must not
    change input Series/DataFrame (though pandas doesn't check it).
inplace : bool, default False
    Whether to perform the operation in place on the data.
axis : int, default None
    Alignment axis if needed.
level : int, default None
    Alignment level if needed.
errors : str, {{'raise', 'ignore'}}, default 'raise'
    Note that currently this parameter won't affect
    the results and will always coerce to a suitable dtype.

    - 'raise' : allow exceptions to be raised.
    - 'ignore' : suppress exceptions. On error return original object.

try_cast : bool, default False
    Try to cast the result back to the input type (if possible).

Returns
-------
Same type as caller

See Also
--------
:func:`DataFrame.{opposite}` : Return an object of same shape as
    self.

Notes
-----
The mask method is an application of the if-then idiom. For each
element in the calling DataFrame, if ``cond`` is ``False`` the
element is used; otherwise the corresponding element from the DataFrame
``other`` is used.

The signature for :func:`DataFrame.where` differs from
:func:`numpy.where`. Roughly ``df1.where(m, df2)`` is equivalent to
``np.where(m, df1, df2)``.

For further details and examples see the ``mask`` documentation in
:ref:`indexing <indexing.where_mask>`.

Examples
--------
>>> import mars.tensor as mt
>>> import mars.dataframe as md
>>> s = md.Series(range(5))
>>> s.where(s > 0).execute()
0    NaN
1    1.0
2    2.0
3    3.0
4    4.0
dtype: float64

>>> s.mask(s > 0).execute()
0    0.0
1    NaN
2    NaN
3    NaN
4    NaN
dtype: float64

>>> s.where(s > 1, 10).execute()
0    10
1    10
2    2
3    3
4    4
dtype: int64

>>> df = md.DataFrame(mt.arange(10).reshape(-1, 2), columns=['A', 'B'])
>>> df.execute()
   A  B
0  0  1
1  2  3
2  4  5
3  6  7
4  8  9
>>> m = df % 3 == 0
>>> df.where(m, -df).execute()
   A  B
0  0 -1
1 -2  3
2 -4 -5
3  6 -7
4 -8  9
>>> df.where(m, -df) == mt.where(m, df, -df).execute()
      A     B
0  True  True
1  True  True
2  True  True
3  True  True
4  True  True
>>> df.where(m, -df) == df.mask(~m, -df).execute()
      A     B
0  True  True
1  True  True
2  True  True
3  True  True
4  True  True
"""


def _where(
    df_or_series,
    cond,
    other=np.nan,
    inplace=False,
    axis=None,
    level=None,
    errors="raise",
    try_cast=False,
    replace_true=False,
):
    if df_or_series.ndim == 2 and getattr(other, "ndim", 2) == 1 and axis is None:
        raise ValueError("Must specify axis=0 or 1")

    axis = validate_axis(axis or 0, df_or_series)
    op = DataFrameWhere(
        cond=cond,
        other=other,
        axis=axis,
        level=level,
        errors=errors,
        try_cast=try_cast,
        replace_true=replace_true,
    )
    result = op(df_or_series)
    if inplace:
        df_or_series.data = result.data
    else:
        return result


def where(
    df_or_series,
    cond,
    other=np.nan,
    inplace=False,
    axis=None,
    level=None,
    errors="raise",
    try_cast=False,
):
    return _where(
        df_or_series,
        cond,
        other=other,
        inplace=inplace,
        axis=axis,
        level=level,
        errors=errors,
        try_cast=try_cast,
        replace_true=False,
    )


def mask(
    df_or_series,
    cond,
    other=np.nan,
    inplace=False,
    axis=None,
    level=None,
    errors="raise",
    try_cast=False,
):
    return _where(
        df_or_series,
        cond,
        other=other,
        inplace=inplace,
        axis=axis,
        level=level,
        errors=errors,
        try_cast=try_cast,
        replace_true=True,
    )


mask.__doc__ = _doc_template.format(replace_true=True, opposite="where")
where.__doc__ = _doc_template.format(replace_true=False, opposite="mask")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import scipy.sparse as sps
    import scipy.sparse.linalg as splinalg
except ImportError:  # pragma: no cover
    sps = None
    splinalg = None

from ...utils import lazy_import

splinalg = splinalg
cp = lazy_import("cupy", rename="cp")
cps = lazy_import("cupy.sparse", rename="cps")


def issparse(x):
    if cps and cps.issparse(x):
        # is cupy.sparse
        return True
    if sps and sps.issparse(x):
        # is scipy.sparse
        return True
    if np and isinstance(x, np.ndarray):
        return False
    if cp and isinstance(x, cp.ndarray):
        return False

    from .array import SparseNDArray

    return isinstance(x, SparseNDArray)


def is_sparse_or_dense(x):
    if issparse(x):
        return True
    m = get_array_module(x)
    if m.isscalar(x):
        return True
    return isinstance(x, m.ndarray)


def get_dense_module(x):
    from .array import SparseNDArray

    if cp:
        if isinstance(x, SparseNDArray):
            return get_array_module(x.raw)
        return get_array_module(x)

    return np


def get_array_module(x):
    if cp:
        return cp.get_array_module(x)
    return np


def get_sparse_module(x):
    m = get_array_module(x)
    if m is np:
        return sps
    return cps


def is_cupy(x):
    return get_array_module(x) is cp


def naked(x):
    if hasattr(x, "spmatrix"):
        return x.spmatrix
    if not is_sparse_or_dense(x):
        raise TypeError("only sparse matrix or ndarray accepted")
    return x



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urlparse

from ...utils import parse_readable_size, calc_size_by_str
from ..kubernetes.config import ContainerEnvConfig

DEFAULT_SERVICE_ACCOUNT_NAME = "kubedl-sa"


def _remove_nones(cfg):
    return dict((k, v) for k, v in cfg.items() if v is not None)


class ResourceConfig:
    """
    Configuration builder for Kubernetes computation resources
    """

    def __init__(self, cpu, memory):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

    def build(self):
        return {
            "cpu": str(self._cpu),
            "memory": str(int(self._memory)),
        }


class ReplicaSpecConfig:
    """
    Base configuration builder for Kubernetes replication controllers
    """

    container_name = "mars"

    def __init__(
        self,
        name,
        image,
        replicas,
        resource_request=None,
        resource_limit=None,
        node_selectors=None,
    ):
        self._name = name
        self._image = image
        self._replicas = replicas
        self._envs = dict()
        self._node_selectors = node_selectors

        self.add_default_envs()

        self._resource_request = resource_request
        self._resource_limit = resource_limit

    def add_env(self, name, value=None, field_path=None):
        self._envs[name] = ContainerEnvConfig(name, value=value, field_path=field_path)

    def add_simple_envs(self, envs):
        for k, v in envs.items() or ():
            self.add_env(k, v)

    def add_default_envs(self):
        pass  # pragma: no cover

    def build_container_command(self):
        raise NotImplementedError

    def build_container(self):
        resources_dict = {
            "requests": self._resource_request.build()
            if self._resource_request
            else None,
            "limits": self._resource_limit.build() if self._resource_limit else None,
        }
        return _remove_nones(
            {
                "imagePullPolicy": "Always",
                "command": self.build_container_command(),
                "env": [env.build() for env in self._envs.values()] or None,
                "image": self._image,
                "name": self.container_name,
                "resources": dict((k, v) for k, v in resources_dict.items() if v)
                or None,
            }
        )

    def build_template_spec(self):
        return _remove_nones(
            {
                "serviceAccountName": DEFAULT_SERVICE_ACCOUNT_NAME,
                "nodeSelector": self._node_selectors,
                "containers": [self.build_container()],
            }
        )

    def build(self):
        return {
            "replicas": int(self._replicas),
            "restartPolicy": "Never",
            "template": {
                "metadata": {
                    "labels": {"mars/service-type": self._name},
                },
                "spec": self.build_template_spec(),
            },
        }


class MarsReplicaSpecConfig(ReplicaSpecConfig):
    service_name = None
    service_label = None

    def __init__(
        self,
        image,
        replicas,
        cpu=None,
        memory=None,
        limit_resources_ratio=1.2,
        memory_limit_ratio=2,
        modules=None,
        node_selectors=None,
    ):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

        if isinstance(modules, str):
            self._modules = modules.split(",")
        else:
            self._modules = modules

        res_request = ResourceConfig(cpu, memory) if cpu or memory else None
        memory_limit_ratio = (
            memory_limit_ratio
            if memory_limit_ratio is not None
            else limit_resources_ratio
        )
        res_limit = (
            ResourceConfig(cpu * limit_resources_ratio, memory * memory_limit_ratio)
            if cpu or memory
            else None
        )
        super().__init__(
            self.service_label,
            image,
            replicas,
            resource_request=res_request,
            resource_limit=res_limit,
            node_selectors=node_selectors,
        )

    def build_container_command(self):
        cmd = [
            "/srv/entrypoint.sh",
            f"mars.deploy.kubernetes.{self.service_name}",
        ]
        return cmd

    def add_default_envs(self):
        if self._cpu:
            self.add_env("MARS_CPU_TOTAL", str(self._cpu))

        if self._memory:
            self.add_env("MARS_MEMORY_TOTAL", str(int(self._memory)))

        if self._modules:
            self.add_env("MARS_LOAD_MODULES", ",".join(self._modules))


class MarsSchedulerSpecConfig(MarsReplicaSpecConfig):
    service_name = "scheduler"
    service_label = "marsscheduler"


class MarsWorkerSpecConfig(MarsReplicaSpecConfig):
    service_name = "worker"
    service_label = "marsworker"

    def __init__(self, *args, **kwargs):
        cache_mem = kwargs.pop("cache_mem", None)
        self._spill_dirs = kwargs.pop("spill_dirs", None) or ()
        # set limits as 2*requests for worker replica defaulted.
        kwargs["limit_resources_ratio"] = kwargs.get("limit_resources_ratio", 1.2)
        super().__init__(*args, **kwargs)
        self._cache_mem = calc_size_by_str(cache_mem, self._memory)
        self.add_env("MARS_CACHE_MEM_SIZE", self._cache_mem)

    @property
    def spill_dirs(self):
        return self._spill_dirs

    @property
    def cache_mem(self):
        return self._cache_mem

    def add_default_envs(self):
        super().add_default_envs()
        if self._spill_dirs:
            self.add_env("MARS_SPILL_DIRS", ":".join(self._spill_dirs))


class MarsWebSpecConfig(MarsReplicaSpecConfig):
    service_name = "web"
    service_label = "marsweb"


class MarsJobConfig:
    def __init__(
        self, job_name, scheduler_config, worker_config, web_config, web_host=None
    ):
        self._job_name = job_name
        self._scheduler_config = scheduler_config
        self._worker_config = worker_config
        self._web_config = web_config
        self._web_host = web_host

    def build(self):
        if self._job_name is None:
            metadata = {"generateName": "mars-job-"}
        else:
            metadata = {"name": self._job_name}

        web_host = self._web_host
        if web_host is not None and "://" in web_host:
            web_host = urlparse(web_host).netloc

        return {
            "kind": "MarsJob",
            "metadata": metadata,
            "spec": _remove_nones(
                {
                    "workerMemoryTuningPolicy": _remove_nones(
                        {
                            "spillDirs": self._worker_config.spill_dirs,
                            "workerCacheSize": self._worker_config.cache_mem,
                        }
                    ),
                    "cleanPodPolicy": "None",
                    "webHost": web_host,
                    "marsReplicaSpecs": {
                        "Worker": self._worker_config.build(),
                        "Scheduler": self._scheduler_config.build(),
                        "WebService": self._web_config.build(),
                    },
                }
            ),
        }



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

from collections import namedtuple, defaultdict
from typing import Dict, List
from ..api import Fetcher, register_fetcher_cls


_GetWithIndex = namedtuple("GetWithIndex", ["get", "index"])


@register_fetcher_cls
class MarsFetcher(Fetcher):
    name = "mars"
    required_meta_keys = ("bands",)

    def __init__(self, get_storage_api, **kwargs):
        self._get_storage_api = get_storage_api
        self._storage_api_to_gets = defaultdict(list)
        self._counter = 0

    async def append(self, chunk_key: str, chunk_meta: Dict, conditions: List = None):
        band = None
        if chunk_meta:
            bands = chunk_meta.get("bands")
            if bands:
                band = bands[0]
        storage_api = await self._get_storage_api(band)
        get = _GetWithIndex(
            storage_api.get.delay(chunk_key, conditions=conditions), self._counter
        )
        self._storage_api_to_gets[storage_api].append(get)
        self._counter += 1

    async def get(self):
        results = [None] * self._counter
        for storage_api in self._storage_api_to_gets:
            gets = self._storage_api_to_gets[storage_api]
            fetched_data = await storage_api.get.batch(
                *map(operator.itemgetter(0), gets)
            )
            for get, data in zip(gets, fetched_data):
                results[get.index] = data
        return results



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import platform
import socket
import sys
from typing import Dict, List

import numpy as np
import pandas as pd

from ...resource import Resource, ZeroResource

try:
    import scipy
except ImportError:  # pragma: no cover
    scipy = None

from ... import resource as mars_resource
from ...config import options
from ...utils import git_info, lazy_import
from ...storage import StorageLevel
from .core import WorkerSlotInfo, QuotaInfo, DiskInfo, StorageInfo

cp = lazy_import("cupy", rename="cp")
cudf = lazy_import("cudf")

logger = logging.getLogger(__name__)

_is_initial = True


def gather_node_env():
    from ...lib.mkl_interface import mkl_get_version
    from ...lib.nvutils import NVError
    from ... import __version__ as mars_version

    global _is_initial
    if _is_initial:
        _is_initial = False
        mars_resource.cpu_percent()

    mem_stats = mars_resource.virtual_memory()

    node_info = {
        "command_line": sys.argv,
        "platform": platform.platform(),
        "host_name": socket.gethostname(),
        "python_version": sys.version,
        "mars_version": mars_version,
        "cpu_total": mars_resource.cpu_count(),
        "memory_total": mem_stats.total,
        "options": options.to_dict(),
    }

    if "MARS_K8S_POD_NAME" in os.environ:
        node_info["k8s_pod_name"] = os.environ["MARS_K8S_POD_NAME"]
    if "CONTAINER_ID" in os.environ:
        node_info["yarn_container_id"] = os.environ["CONTAINER_ID"]

    try:
        cuda_info = mars_resource.cuda_info()
    except NVError:  # pragma: no cover
        logger.exception("NVError encountered, cannot gather CUDA devices.")
        cuda_info = None

    if cuda_info:
        node_info["cuda_info"] = {
            "driver": cuda_info.driver_version,
            "cuda": cuda_info.cuda_version,
            "products": list(cuda_info.products),
        }

    package_vers = {
        "numpy": np.__version__,
        "pandas": pd.__version__,
    }
    if hasattr(np, "__mkl_version__") and mkl_get_version:
        mkl_version = mkl_get_version()
        package_vers[
            "mkl"
        ] = f"{mkl_version.major}.{mkl_version.minor}.{mkl_version.update}"

    if scipy is not None:
        package_vers["scipy"] = scipy.__version__
    if cp is not None:
        package_vers["cupy"] = cp.__version__
    if cudf is not None:
        package_vers["cudf"] = cudf.__version__

    node_info["package_versions"] = package_vers

    git = git_info()
    if git:
        node_info["git_info"] = {
            "hash": git.commit_hash,
            "ref": git.commit_ref,
        }

    bands = node_info["bands"] = dict()

    cpu_band = {
        "resources": {
            "cpu": mars_resource.cpu_count(),
            "memory": mars_resource.virtual_memory().total,
        }
    }
    # todo numa can be supported by adding more bands
    bands["numa-0"] = cpu_band

    for idx, gpu_card_stat in enumerate(
        mars_resource.cuda_card_stats()
    ):  # pragma: no cover
        bands[f"gpu-{idx}"] = {
            "resources": {
                "gpu": 1,
                "memory": gpu_card_stat.fb_mem_info.total,
            }
        }
    return node_info


def gather_node_resource(band_to_resource: Dict[str, Resource] = None, use_gpu=True):
    # todo numa can be supported by adding more bands
    res = dict()
    mem_info = mars_resource.virtual_memory()
    num_cpu = (
        mars_resource.cpu_count()
        if band_to_resource is None
        else band_to_resource.get("numa-0", ZeroResource).num_cpus
    )
    mem_bytes = (
        mem_info.total
        if band_to_resource is None
        else band_to_resource.get("numa-0", ZeroResource).mem_bytes
    )
    if num_cpu:  # pragma: no branch
        res["numa-0"] = {
            "cpu_avail": mars_resource.cpu_count()
            - mars_resource.cpu_percent() / 100.0,
            "cpu_total": num_cpu,
            "memory_avail": mem_info.available,
            "memory_total": min(mem_info.total, mem_bytes),
        }

    if use_gpu:
        for idx, gpu_card_stat in enumerate(
            mars_resource.cuda_card_stats()
        ):  # pragma: no cover
            num_gpu = (
                1
                if band_to_resource is None
                else band_to_resource.get(f"gpu-{idx}", ZeroResource).num_gpus
            )
            if not num_gpu:
                continue
            res[f"gpu-{idx}"] = {
                "gpu_avail": 1 - gpu_card_stat.gpu_usage,
                "gpu_total": num_gpu,
                "memory_avail": gpu_card_stat.fb_mem_info.available,
                "memory_total": gpu_card_stat.fb_mem_info.total,
            }
    return res


def gather_node_details(
    band_slot_infos: Dict[str, List[WorkerSlotInfo]] = None,
    band_quota_infos: Dict[str, QuotaInfo] = None,
    disk_infos: List[DiskInfo] = None,
    band_storage_infos: Dict[str, Dict[StorageLevel, StorageInfo]] = None,
):
    disk_io_usage = mars_resource.disk_io_usage()
    net_io_usage = mars_resource.net_io_usage()
    res = {
        "disk": dict(zip(("reads", "writes"), disk_io_usage))
        if disk_io_usage
        else dict(),
        "network": dict(zip(("receives", "sends"), net_io_usage))
        if net_io_usage
        else dict(),
        "iowait": mars_resource.iowait(),
    }

    if disk_infos:
        part_dict = dict()
        for info in disk_infos:
            part_dev = mars_resource.get_path_device(info.path)
            if part_dev in part_dict:
                continue

            disk_usage_result = mars_resource.disk_usage(info.path)
            io_usage_result = mars_resource.disk_io_usage(info.path)
            part_dict[part_dev] = disk_info = {
                "size_limit": info.limit_size,
                "size_used": disk_usage_result.used,
                "size_total": disk_usage_result.total,
            }
            if io_usage_result is not None:
                disk_info.update(
                    {
                        "reads": io_usage_result.reads if io_usage_result else None,
                        "writes": io_usage_result.writes if io_usage_result else None,
                    }
                )
            if not sys.platform.startswith("win"):
                in_usage_result = os.statvfs(info.path)
                disk_info.update(
                    {
                        "inode_used": in_usage_result.f_files
                        - in_usage_result.f_favail,
                        "inode_total": in_usage_result.f_files,
                    }
                )
        res["disk"]["partitions"] = part_dict

    band_slot_infos = band_slot_infos or dict()
    res["slot"] = {
        band: [
            {
                "slot_id": slot_info.slot_id,
                "session_id": slot_info.session_id,
                "subtask_id": slot_info.subtask_id,
                "processor_usage": slot_info.processor_usage,
            }
            for slot_info in slot_infos
        ]
        for band, slot_infos in band_slot_infos.items()
    }

    band_quota_infos = band_quota_infos or dict()
    res["quota"] = {
        band: {
            "quota_size": quota_info.quota_size,
            "allocated_size": quota_info.allocated_size,
            "hold_size": quota_info.hold_size,
        }
        for band, quota_info in band_quota_infos.items()
    }

    band_storage_infos = band_storage_infos or dict()
    res["storage"] = {
        band: {
            level.name.lower(): {
                "size_used": storage_info.used_size,
                "size_total": storage_info.total_size,
                "size_pinned": storage_info.pinned_size,
            }
            for level, storage_info in storage_infos.items()
        }
        for band, storage_infos in band_storage_infos.items()
    }
    return res



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from functools import partial

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, Entity, OutputType
from ...core.operand import OperandStage, MapReduceOperand
from ...lib.groupby_wrapper import wrapped_groupby
from ...serialization.serializables import BoolField, Int32Field, AnyField
from ...utils import lazy_import
from ..align import align_dataframe_series, align_series_series
from ..initializer import Series as asseries
from ..core import SERIES_TYPE, SERIES_CHUNK_TYPE
from ..utils import (
    build_concatenated_rows_frame,
    hash_dataframe_on,
    build_df,
    build_series,
    parse_index,
    is_cudf,
)
from ..operands import DataFrameOperandMixin, DataFrameShuffleProxy


cudf = lazy_import("cudf")


class DataFrameGroupByOperand(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.GROUPBY

    _by = AnyField("by", on_serialize=lambda x: x.data if isinstance(x, Entity) else x)
    _level = AnyField("level")
    _as_index = BoolField("as_index")
    _sort = BoolField("sort")
    _group_keys = BoolField("group_keys")

    _shuffle_size = Int32Field("shuffle_size")

    def __init__(
        self,
        by=None,
        level=None,
        as_index=None,
        sort=None,
        group_keys=None,
        shuffle_size=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _by=by,
            _level=level,
            _as_index=as_index,
            _sort=sort,
            _group_keys=group_keys,
            _shuffle_size=shuffle_size,
            _output_types=output_types,
            **kw
        )
        if output_types:
            if self.stage in (OperandStage.map, OperandStage.reduce):
                if output_types[0] in (
                    OutputType.dataframe,
                    OutputType.dataframe_groupby,
                ):
                    output_types = [OutputType.dataframe]
                else:
                    output_types = [OutputType.series]
            else:
                if output_types[0] in (
                    OutputType.dataframe,
                    OutputType.dataframe_groupby,
                ):
                    output_types = [OutputType.dataframe_groupby]
                elif output_types[0] == OutputType.series:
                    output_types = [OutputType.series_groupby]
            self.output_types = output_types

    @property
    def by(self):
        return self._by

    @property
    def level(self):
        return self._level

    @property
    def as_index(self):
        return self._as_index

    @property
    def sort(self):
        return self._sort

    @property
    def group_keys(self):
        return self._group_keys

    @property
    def shuffle_size(self):
        return self._shuffle_size

    @property
    def is_dataframe_obj(self):
        return self.output_types[0] in (
            OutputType.dataframe_groupby,
            OutputType.dataframe,
        )

    @property
    def groupby_params(self):
        return dict(
            by=self.by,
            level=self.level,
            as_index=self.as_index,
            sort=self.sort,
            group_keys=self.group_keys,
        )

    def build_mock_groupby(self, **kwargs):
        in_df = self.inputs[0]
        if self.is_dataframe_obj:
            mock_obj = build_df(
                in_df, size=[2, 2], fill_value=[1, 2], ensure_string=True
            )
        else:
            mock_obj = build_series(
                in_df,
                size=[2, 2],
                fill_value=[1, 2],
                name=in_df.name,
                ensure_string=True,
            )

        new_kw = self.groupby_params
        new_kw.update(kwargs)
        if new_kw.get("level"):
            new_kw["level"] = 0
        if isinstance(new_kw["by"], list):
            new_by = []
            for v in new_kw["by"]:
                if isinstance(v, ENTITY_TYPE):
                    build_fun = build_df if v.ndim == 2 else build_series
                    mock_by = pd.concat(
                        [
                            build_fun(v, size=2, fill_value=1, name=v.name),
                            build_fun(v, size=2, fill_value=2, name=v.name),
                        ]
                    )
                    new_by.append(mock_by)
                else:
                    new_by.append(v)
            new_kw["by"] = new_by
        return mock_obj.groupby(**new_kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        if len(inputs) > 1:
            by = []
            for k in self._by:
                if isinstance(k, (SERIES_TYPE, SERIES_CHUNK_TYPE)):
                    by.append(next(inputs_iter))
                else:
                    by.append(k)
            self._by = by

    def __call__(self, df):
        params = df.params.copy()
        params["index_value"] = parse_index(None, df.key, df.index_value.key)
        if df.ndim == 2:
            if isinstance(self.by, list):
                index, types = [], []
                for k in self.by:
                    if isinstance(k, SERIES_TYPE):
                        index.append(k.name)
                        types.append(k.dtype)
                    elif k in df.dtypes:
                        index.append(k)
                        types.append(df.dtypes[k])
                    else:
                        raise KeyError(k)
                params["key_dtypes"] = pd.Series(types, index=index)

        inputs = [df]
        if isinstance(self.by, list):
            for k in self.by:
                if isinstance(k, SERIES_TYPE):
                    inputs.append(k)

        return self.new_tileable(inputs, **params)

    @classmethod
    def _align_input_and_by(cls, op, inp, by):
        align_method = (
            partial(align_dataframe_series, axis="index")
            if op.is_dataframe_obj
            else align_series_series
        )
        nsplits, _, inp_chunks, by_chunks = align_method(inp, by)

        inp_params = inp.params
        inp_params["chunks"] = inp_chunks
        inp_params["nsplits"] = nsplits
        inp = inp.op.copy().new_tileable(op.inputs, kws=[inp_params])

        by_params = by.params
        by_params["chunks"] = by_chunks
        if len(nsplits) == 2:
            by_nsplits = nsplits[:1]
        else:
            by_nsplits = nsplits
        by_params["nsplits"] = by_nsplits
        by = by.op.copy().new_tileable(by.op.inputs, kws=[by_params])

        return inp, by

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        by = op.by

        series_in_by = False
        new_inputs = []
        if len(op.inputs) > 1:
            # by series
            new_by = []
            for k in by:
                if isinstance(k, SERIES_TYPE):
                    in_df, k = cls._align_input_and_by(op, in_df, k)
                    if len(new_inputs) == 0:
                        new_inputs.append(in_df)
                    new_inputs.append(k)
                    series_in_by = True
                new_by.append(k)
            by = new_by
        else:
            new_inputs = op.inputs

        is_dataframe_obj = op.is_dataframe_obj
        if is_dataframe_obj:
            in_df = build_concatenated_rows_frame(in_df)
            output_type = OutputType.dataframe
            chunk_shape = (in_df.chunk_shape[0], 1)
        else:
            output_type = OutputType.series
            chunk_shape = (in_df.chunk_shape[0],)

        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op.stage = OperandStage.map
            map_op._shuffle_size = chunk_shape[0]
            map_op._output_types = [output_type]
            chunk_inputs = [chunk]
            if len(op.inputs) > 1:
                chunk_by = []
                for k in by:
                    if isinstance(k, SERIES_TYPE):
                        by_chunk = k.cix[
                            chunk.index[0],
                        ]
                        chunk_by.append(by_chunk)
                        chunk_inputs.append(by_chunk)
                    else:
                        chunk_by.append(k)
                map_op._by = chunk_by
            map_chunks.append(
                map_op.new_chunk(
                    chunk_inputs, shape=(np.nan, np.nan), index=chunk.index
                )
            )

        proxy_chunk = DataFrameShuffleProxy(output_types=[output_type]).new_chunk(
            map_chunks, shape=()
        )

        # generate reduce chunks
        reduce_chunks = []
        out_indices = list(itertools.product(*(range(s) for s in chunk_shape)))
        for ordinal, out_idx in enumerate(out_indices):
            reduce_op = op.copy().reset_key()
            reduce_op._by = None
            reduce_op._output_types = [output_type]
            reduce_op.stage = OperandStage.reduce
            reduce_op.reducer_ordinal = ordinal
            reduce_op.n_reducers = len(out_indices)
            reduce_chunks.append(
                reduce_op.new_chunk(
                    [proxy_chunk], shape=(np.nan, np.nan), index=out_idx
                )
            )

        # generate groupby chunks
        out_chunks = []
        for chunk in reduce_chunks:
            groupby_op = op.copy().reset_key()
            if series_in_by:
                # set by to None, cuz data of by will be passed from map to reduce to groupby
                groupby_op._by = None
            if is_dataframe_obj:
                new_shape = (np.nan, in_df.shape[1])
            else:
                new_shape = (np.nan,)
            params = dict(shape=new_shape, index=chunk.index)
            if op.is_dataframe_obj:
                params.update(
                    dict(
                        dtypes=in_df.dtypes,
                        columns_value=in_df.columns_value,
                        index_value=parse_index(None, chunk.key, proxy_chunk.key),
                    )
                )
            else:
                params.update(
                    dict(
                        name=in_df.name,
                        dtype=in_df.dtype,
                        index_value=parse_index(None, chunk.key, proxy_chunk.key),
                    )
                )
            out_chunks.append(groupby_op.new_chunk([chunk], **params))

        new_op = op.copy()
        params = op.outputs[0].params.copy()
        if is_dataframe_obj:
            params["nsplits"] = ((np.nan,) * len(out_chunks), (in_df.shape[1],))
        else:
            params["nsplits"] = ((np.nan,) * len(out_chunks),)
        params["chunks"] = out_chunks
        return new_op.new_tileables(new_inputs, **params)

    @classmethod
    def execute_map(cls, ctx, op):
        is_dataframe_obj = op.is_dataframe_obj
        by = op.by
        chunk = op.outputs[0]
        df = ctx[op.inputs[0].key]

        deliver_by = False  # output by for the upcoming process
        if isinstance(by, list):
            new_by = []
            for v in by:
                if isinstance(v, ENTITY_TYPE):
                    deliver_by = True
                    new_by.append(ctx[v.key])
                else:
                    new_by.append(v)
            by = new_by

        if isinstance(by, list) or callable(by):
            on = by
        else:
            on = None

        if isinstance(df, tuple):
            filters = hash_dataframe_on(df[0], on, op.shuffle_size, level=op.level)
        else:
            filters = hash_dataframe_on(df, on, op.shuffle_size, level=op.level)

        def _take_index(src, f):
            result = src.iloc[f]
            if src.index.names:
                result.index.names = src.index.names
            if isinstance(src.index, pd.MultiIndex):
                result.index = result.index.remove_unused_levels()
            if is_cudf(result):  # pragma: no cover
                result = result.copy()
            return result

        for index_idx, index_filter in enumerate(filters):
            if is_dataframe_obj:
                reducer_index = (index_idx, chunk.index[1])
            else:
                reducer_index = (index_idx,)

            if deliver_by:
                filtered_by = []
                for v in by:
                    if isinstance(v, pd.Series):
                        filtered_by.append(_take_index(v, index_filter))
                    else:
                        filtered_by.append(v)
                if isinstance(df, tuple):
                    ctx[chunk.key, reducer_index] = tuple(
                        _take_index(x, index_filter) for x in df
                    ) + (filtered_by, deliver_by)
                else:
                    ctx[chunk.key, reducer_index] = (
                        _take_index(df, index_filter),
                        filtered_by,
                        deliver_by,
                    )
            else:
                if isinstance(df, tuple):
                    ctx[chunk.key, reducer_index] = tuple(
                        _take_index(x, index_filter) for x in df
                    ) + (deliver_by,)
                else:
                    ctx[chunk.key, reducer_index] = _take_index(df, index_filter)

    @classmethod
    def execute_reduce(cls, ctx, op: "DataFrameGroupByOperand"):
        xdf = cudf if op.gpu else pd
        chunk = op.outputs[0]
        input_idx_to_df = dict(op.iter_mapper_data_with_index(ctx))
        row_idxes = sorted(input_idx_to_df.keys())

        res = []
        for row_idx in row_idxes:
            row_df = input_idx_to_df.get(row_idx, None)
            if row_df is not None:
                res.append(row_df)
        by = None
        if isinstance(res[0], tuple):
            # By is series
            deliver_by = res[0][-1]
            r = []
            part_len = len(res[0])
            part_len -= 1 if not deliver_by else 2
            for n in range(part_len):
                r.append(xdf.concat([it[n] for it in res], axis=0))
            r = tuple(r)

            if deliver_by:
                by = [None] * len(res[0][-2])
                for it in res:
                    for i, v in enumerate(it[1]):
                        if isinstance(v, pd.Series):
                            if by[i] is None:
                                by[i] = v
                            else:
                                by[i] = pd.concat([by[i], v], axis=0)
                        else:
                            by[i] = v
        else:
            r = pd.concat(res, axis=0)

        if chunk.index_value is not None:
            if isinstance(r, tuple):
                for s in r:
                    s.index.name = chunk.index_value.name
            else:
                r.index.name = chunk.index_value.name
        if by is None:
            ctx[chunk.key] = r
        elif isinstance(r, tuple):
            ctx[chunk.key] = r + (by,)
        else:
            ctx[chunk.key] = (r, by)

    @classmethod
    def execute(cls, ctx, op: "DataFrameGroupByOperand"):
        if op.stage == OperandStage.map:
            cls.execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls.execute_reduce(ctx, op)
        else:
            inp = ctx[op.inputs[0].key]
            if isinstance(inp, tuple):
                # df, by
                df, by = inp
            else:
                df = inp
                by = op.by
            ctx[op.outputs[0].key] = wrapped_groupby(
                df,
                by=by,
                level=op.level,
                as_index=op.as_index,
                sort=op.sort,
                group_keys=op.group_keys,
            )


def groupby(df, by=None, level=None, as_index=True, sort=True, group_keys=True):
    if not as_index and df.op.output_types[0] == OutputType.series:
        raise TypeError("as_index=False only valid with DataFrame")

    output_types = (
        [OutputType.dataframe_groupby] if df.ndim == 2 else [OutputType.series_groupby]
    )
    if isinstance(by, (SERIES_TYPE, pd.Series)):
        if isinstance(by, pd.Series):
            by = asseries(by)
        by = [by]
    elif df.ndim > 1 and by is not None and not isinstance(by, list):
        by = [by]
    op = DataFrameGroupByOperand(
        by=by,
        level=level,
        as_index=as_index,
        sort=sort,
        group_keys=group_keys,
        output_types=output_types,
    )
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...serialization.serializables import AnyField, StringField, ListField
from ...tensor.base import sort
from ...utils import pd_release_version
from ..core import DATAFRAME_TYPE, SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, build_empty_series, parse_index

_need_astype_contiguous = pd_release_version == (1, 3, 0)


class DataFrameAstype(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.ASTYPE

    _dtype_values = AnyField("dtype_values")
    _errors = StringField("errors")
    _category_cols = ListField("category_cols")

    def __init__(
        self,
        dtype_values=None,
        errors=None,
        category_cols=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _dtype_values=dtype_values,
            _errors=errors,
            _category_cols=category_cols,
            _output_types=output_types,
            **kw
        )

    @property
    def dtype_values(self):
        return self._dtype_values

    @property
    def errors(self):
        return self._errors

    @property
    def category_cols(self):
        return self._category_cols

    @classmethod
    def _tile_one_chunk(cls, op):
        c = op.inputs[0].chunks[0]
        chunk_op = op.copy().reset_key()
        chunk_params = op.outputs[0].params.copy()
        chunk_params["index"] = c.index
        out_chunks = [chunk_op.new_chunk([c], **chunk_params)]

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            nsplits=op.inputs[0].nsplits,
            chunks=out_chunks,
            **op.outputs[0].params.copy()
        )

    @classmethod
    def _tile_series_index(cls, op):
        in_series = op.inputs[0]
        out = op.outputs[0]

        unique_chunk = None
        if op.dtype_values == "category" and isinstance(op.dtype_values, str):
            unique_chunk = (yield from recursive_tile(sort(in_series.unique()))).chunks[
                0
            ]

        chunks = []
        for c in in_series.chunks:
            chunk_op = op.copy().reset_key()
            params = c.params.copy()
            params["dtype"] = out.dtype
            if unique_chunk is not None:
                chunk_op._category_cols = [in_series.name]
                new_chunk = chunk_op.new_chunk([c, unique_chunk], **params)
            else:
                new_chunk = chunk_op.new_chunk([c], **params)
            chunks.append(new_chunk)

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs, nsplits=in_series.nsplits, chunks=chunks, **out.params.copy()
        )

    @classmethod
    def _tile_dataframe(cls, op):
        in_df = op.inputs[0]
        out = op.outputs[0]
        cum_nsplits = np.cumsum((0,) + in_df.nsplits[1])
        out_chunks = []

        if op.dtype_values == "category":
            # all columns need unique values
            for c in in_df.chunks:
                chunk_op = op.copy().reset_key()
                params = c.params.copy()
                dtypes = out.dtypes[
                    cum_nsplits[c.index[1]] : cum_nsplits[c.index[1] + 1]
                ]
                params["dtypes"] = dtypes
                chunk_op._category_cols = list(c.columns_value.to_pandas())
                unique_chunks = []
                for col in c.columns_value.to_pandas():
                    unique = yield from recursive_tile(sort(in_df[col].unique()))
                    unique_chunks.append(unique.chunks[0])
                new_chunk = chunk_op.new_chunk([c] + unique_chunks, **params)
                out_chunks.append(new_chunk)
        elif (
            isinstance(op.dtype_values, dict) and "category" in op.dtype_values.values()
        ):
            # some columns' types are category
            category_cols = [
                c
                for c, v in op.dtype_values.items()
                if isinstance(v, str) and v == "category"
            ]
            unique_chunks = dict()
            for col in category_cols:
                unique = yield from recursive_tile(sort(in_df[col].unique()))
                unique_chunks[col] = unique.chunks[0]
            for c in in_df.chunks:
                chunk_op = op.copy().reset_key()
                params = c.params.copy()
                dtypes = out.dtypes[
                    cum_nsplits[c.index[1]] : cum_nsplits[c.index[1] + 1]
                ]
                params["dtypes"] = dtypes
                chunk_category_cols = []
                chunk_unique_chunks = []
                for col in c.columns_value.to_pandas():
                    if col in category_cols:
                        chunk_category_cols.append(col)
                        chunk_unique_chunks.append(unique_chunks[col])
                chunk_op._category_cols = chunk_category_cols
                new_chunk = chunk_op.new_chunk([c] + chunk_unique_chunks, **params)
                out_chunks.append(new_chunk)
        else:
            for c in in_df.chunks:
                chunk_op = op.copy().reset_key()
                params = c.params.copy()
                dtypes = out.dtypes[
                    cum_nsplits[c.index[1]] : cum_nsplits[c.index[1] + 1]
                ]
                params["dtypes"] = dtypes
                new_chunk = chunk_op.new_chunk([c], **params)
                out_chunks.append(new_chunk)

        new_op = op.copy()
        return new_op.new_dataframes(
            op.inputs, nsplits=in_df.nsplits, chunks=out_chunks, **out.params.copy()
        )

    @classmethod
    def tile(cls, op):
        if len(op.inputs[0].chunks) == 1:
            return cls._tile_one_chunk(op)
        elif isinstance(op.inputs[0], DATAFRAME_TYPE):
            return (yield from cls._tile_dataframe(op))
        else:
            return (yield from cls._tile_series_index(op))

    @classmethod
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        if not isinstance(op.dtype_values, dict):
            if op.category_cols is not None:
                uniques = [ctx[c.key] for c in op.inputs[1:]]
                dtype = dict(
                    (col, CategoricalDtype(unique_values))
                    for col, unique_values in zip(op.category_cols, uniques)
                )
                ctx[op.outputs[0].key] = in_data.astype(dtype, errors=op.errors)

            elif isinstance(in_data, pd.Index):
                ctx[op.outputs[0].key] = in_data.astype(op.dtype_values)
            else:
                if _need_astype_contiguous and not in_data.values.flags.contiguous:
                    # astype changes the data order in pandas==1.3.0, see pandas#42396
                    in_data = in_data.copy()
                ctx[op.outputs[0].key] = in_data.astype(
                    op.dtype_values, errors=op.errors
                )
        else:
            selected_dtype = dict(
                (k, v) for k, v in op.dtype_values.items() if k in in_data.columns
            )
            if op.category_cols is not None:
                uniques = [ctx[c.key] for c in op.inputs[1:]]
                for col, unique_values in zip(op.category_cols, uniques):
                    selected_dtype[col] = CategoricalDtype(unique_values)
            ctx[op.outputs[0].key] = in_data.astype(selected_dtype, errors=op.errors)

    def __call__(self, df):
        if isinstance(df, DATAFRAME_TYPE):
            empty_df = build_empty_df(df.dtypes)
            new_df = empty_df.astype(self.dtype_values, errors=self.errors)
            dtypes = []
            for dt, new_dt in zip(df.dtypes, new_df.dtypes):
                if new_dt != dt and isinstance(new_dt, CategoricalDtype):
                    dtypes.append(CategoricalDtype())
                else:
                    dtypes.append(new_dt)
            dtypes = pd.Series(dtypes, index=new_df.dtypes.index)
            return self.new_dataframe(
                [df],
                shape=df.shape,
                dtypes=dtypes,
                index_value=df.index_value,
                columns_value=df.columns_value,
            )
        else:
            empty_series = build_empty_series(df.dtype)
            new_series = empty_series.astype(self.dtype_values, errors=self.errors)
            if new_series.dtype != df.dtype:
                dtype = (
                    CategoricalDtype()
                    if isinstance(new_series.dtype, CategoricalDtype)
                    else new_series.dtype
                )
            else:  # pragma: no cover
                dtype = df.dtype

            if isinstance(df, SERIES_TYPE):
                return self.new_series(
                    [df],
                    shape=df.shape,
                    dtype=dtype,
                    name=df.name,
                    index_value=df.index_value,
                )
            else:
                new_index = df.index_value.to_pandas().astype(self.dtype_values)
                new_index_value = parse_index(
                    new_index, store_data=df.index_value.has_value()
                )
                return self.new_index(
                    [df],
                    shape=df.shape,
                    dtype=dtype,
                    name=df.name,
                    index_value=new_index_value,
                )


def astype(df, dtype, copy=True, errors="raise"):
    """
    Cast a pandas object to a specified dtype ``dtype``.

    Parameters
    ----------
    dtype : data type, or dict of column name -> data type
        Use a numpy.dtype or Python type to cast entire pandas object to
        the same type. Alternatively, use {col: dtype, ...}, where col is a
        column label and dtype is a numpy.dtype or Python type to cast one
        or more of the DataFrame's columns to column-specific types.
    copy : bool, default True
        Return a copy when ``copy=True`` (be very careful setting
        ``copy=False`` as changes to values then may propagate to other
        pandas objects).
    errors : {'raise', 'ignore'}, default 'raise'
        Control raising of exceptions on invalid data for provided dtype.

        - ``raise`` : allow exceptions to be raised
        - ``ignore`` : suppress exceptions. On error return original object.

    Returns
    -------
    casted : same type as caller

    See Also
    --------
    to_datetime : Convert argument to datetime.
    to_timedelta : Convert argument to timedelta.
    to_numeric : Convert argument to a numeric type.
    numpy.ndarray.astype : Cast a numpy array to a specified type.

    Examples
    --------
    Create a DataFrame:

    >>> import mars.dataframe as md
    >>> df = md.DataFrame(pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]}))
    >>> df.dtypes
    col1    int64
    col2    int64
    dtype: object

    Cast all columns to int32:

    >>> df.astype('int32').dtypes
    col1    int32
    col2    int32
    dtype: object

    Cast col1 to int32 using a dictionary:

    >>> df.astype({'col1': 'int32'}).dtypes
    col1    int32
    col2    int64
    dtype: object

    Create a series:

    >>> ser = md.Series(pd.Series([1, 2], dtype='int32'))
    >>> ser.execute()
    0    1
    1    2
    dtype: int32
    >>> ser.astype('int64').execute()
    0    1
    1    2
    dtype: int64

    Convert to categorical type:

    >>> ser.astype('category').execute()
    0    1
    1    2
    dtype: category
    Categories (2, int64): [1, 2]

    Convert to ordered categorical type with custom ordering:

    >>> cat_dtype = pd.api.types.CategoricalDtype(
    ...     categories=[2, 1], ordered=True)
    >>> ser.astype(cat_dtype).execute()
    0    1
    1    2
    dtype: category
    Categories (2, int64): [2 < 1]

    Note that using ``copy=False`` and changing data on a new
    pandas object may propagate changes:

    >>> s1 = md.Series(pd.Series([1, 2]))
    >>> s2 = s1.astype('int64', copy=False)
    >>> s1.execute()  # note that s1[0] has changed too
    0     1
    1     2
    dtype: int64
    """
    if isinstance(dtype, dict):
        keys = list(dtype.keys())
        if isinstance(df, SERIES_TYPE):
            if len(keys) != 1 or keys[0] != df.name:
                raise KeyError(
                    "Only the Series name can be used for the key in Series dtype mappings."
                )
            else:
                dtype = list(dtype.values())[0]
        else:
            for k in keys:
                columns = df.columns_value.to_pandas()
                if k not in columns:
                    raise KeyError(
                        "Only a column name can be used for the key in a dtype mappings argument."
                    )
    op = DataFrameAstype(dtype_values=dtype, errors=errors)
    r = op(df)
    if not copy:
        df.data = r.data
        return df
    else:
        return r


def index_astype(ix, dtype, copy=True):
    """
    Create an Index with values cast to dtypes.

    The class of a new Index is determined by dtype. When conversion is
    impossible, a ValueError exception is raised.

    Parameters
    ----------
    dtype : numpy dtype or pandas type
        Note that any signed integer `dtype` is treated as ``'int64'``,
        and any unsigned integer `dtype` is treated as ``'uint64'``,
        regardless of the size.
    copy : bool, default True
        By default, astype always returns a newly allocated object.
        If copy is set to False and internal requirements on dtype are
        satisfied, the original data is used to create a new Index
        or the original Index is returned.

    Returns
    -------
    Index
        Index with values cast to specified dtype.
    """
    return astype(ix, dtype, copy=copy)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bisect
from collections import defaultdict
import sys
from typing import List, Union

import numpy as np

from .... import oscar as mo
from ....typing import ChunkType


class MutableTensorChunkActor(mo.Actor):
    def __init__(
        self,
        session_id: str,
        manager_address: str,
        chunks: List,
        dtype: Union[np.dtype, str],
        default_value: Union[int, float] = 0,
    ) -> None:
        self._session_id = session_id
        self._manager_address = manager_address
        self._chunks = chunks
        self._dtype = dtype
        self._default_value = default_value

        self._storage_api = None
        self._meta_api = None

        self._index_to_chunk = None

    @classmethod
    def gen_uid(cls, session_id: str, name: str, index: int):
        return f"mutable-tensor-chunk-{session_id}-{name}-{index}"

    async def __post_create__(self):
        from ...storage import StorageAPI
        from ...meta import MetaAPI

        self._storage_api = await StorageAPI.create(self._session_id, self.address)
        self._meta_api = await MetaAPI.create(self._session_id, self._manager_address)

        self._index_to_chunk = {
            chunk.index: MutableTensorChunk(
                chunk,
                self._manager_address,
                self.address,
                default_value=self._default_value,
            )
            for chunk in self._chunks
        }

    async def write(self, chunk_index, records):
        chunk: MutableTensorChunk = self._index_to_chunk[chunk_index]
        await chunk.write(records)

    async def read(self, chunk_index, records, chunk_value_shape, timestamp):
        chunk: MutableTensorChunk = self._index_to_chunk[chunk_index]
        return await chunk.read(records, chunk_value_shape, timestamp)

    async def seal(self, timestamp):
        for _, chunk in self._index_to_chunk.items():
            chunk_data = await chunk.seal(timestamp)
            await self._storage_api.put(chunk.chunk.key, chunk_data)
            await self._meta_api.set_chunk_meta(
                chunk.chunk, bands=[(self.address, "numa-0")]
            )


class MutableTensorChunk:
    def __init__(
        self,
        chunk: ChunkType,
        manager_address: str,
        worker_address: str,
        default_value: Union[int, float] = 0,
    ) -> None:
        self._chunk = chunk
        self._manager_address = manager_address
        self._worker_address = worker_address
        self._default_value = default_value

        self._records = defaultdict(list)

    @property
    def chunk(self):
        return self._chunk

    async def write(self, records):
        for flat_index, value, ts in records:
            self._records[flat_index].append((ts, value))

    async def read(self, records, chunk_value_shape, timestamp):
        result = np.full(shape=chunk_value_shape, fill_value=self._default_value)
        for flat_index, value_index in records:
            if flat_index not in self._records:
                continue
            # Find the newest one.
            #
            # FIXME Python doesn't have things like SortedDict or SortedList,
            # we trigger a `sorted` here to ensure the correct semantic and try
            # to be as efficient as possible.
            self._records[flat_index].sort()
            # bitsect will compare on first element in the tuple.
            index = bisect.bisect_right(
                self._records[flat_index], (timestamp, sys.float_info.max)
            )
            if index == 0:
                continue
            result[value_index] = self._records[flat_index][index - 1][
                1
            ]  # take the value
        return result

    async def seal(self, timestamp):
        result = np.full(self._chunk.shape, self._default_value)
        for flat_index, values in self._records.items():
            if flat_index not in self._records:
                continue
            # compute value
            values.sort()
            index = bisect.bisect_right(values, (timestamp, sys.float_info.max))
            if index == 0:
                continue
            # compute value index
            value_index = np.unravel_index(flat_index, self._chunk.shape)
            result[value_index] = values[index - 1][1]  # take the value
        return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import pandas as pd

from ...core import ENTITY_TYPE, ExecutableTuple
from ...utils import adapt_mars_docstring


class PlotAccessor:
    def __init__(self, obj):
        self._obj = obj

    def __call__(self, kind="line", session=None, **kwargs):
        to_executes = OrderedDict()
        to_executes["__object__"] = self._obj

        for k, v in kwargs.items():
            if isinstance(v, ENTITY_TYPE):
                to_executes[k] = v

        result = dict()
        executed = ExecutableTuple(to_executes.values()).execute().fetch()
        for p, v in zip(to_executes, executed):
            result[p] = v

        data = result.pop("__object__")
        pd_kwargs = kwargs.copy()
        pd_kwargs["kind"] = kind
        pd_kwargs.update(result)

        return data.plot(**pd_kwargs)

    @classmethod
    def _gen_func(cls, name, doc):
        def _inner(self, *args, **kwargs):
            return self(kind=name, *args, **kwargs)

        _inner.__name__ = name
        _inner.__doc__ = doc

        return _inner

    @classmethod
    def _register(cls, method):
        doc = getattr(pd.DataFrame.plot, method).__doc__
        new_doc = adapt_mars_docstring(doc)
        setattr(cls, method, cls._gen_func(method, new_doc))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from numbers import Integral

import numpy as np

from ... import opcodes as OperandDef
from ...core import recursive_tile
from ...serialization.serializables import KeyField, AnyField, Int32Field
from ...utils import has_unknown_shape
from ..core import Tensor, TENSOR_TYPE, TENSOR_CHUNK_TYPE, TensorOrder
from ..utils import broadcast_shape, unify_chunks
from ..operands import TensorHasInput, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from .ravel import ravel


class TensorRepeat(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.REPEAT

    _input = KeyField("input")
    _repeats = AnyField("repeats")
    _axis = Int32Field("axis")

    def __init__(self, axis=None, dtype=None, sparse=False, **kw):
        super().__init__(_axis=axis, dtype=dtype, sparse=sparse, **kw)

    @property
    def repeats(self):
        return self._repeats

    @property
    def axis(self):
        return self._axis

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(inputs) > 1:
            self._repeats = self._inputs[1]

    def __call__(self, a, repeats):
        axis = self._axis
        a = astensor(a)
        if axis is None:
            a = ravel(a)

        ax = axis or 0

        if not isinstance(repeats, Integral):
            if not isinstance(repeats, Tensor):
                repeats = np.asarray(repeats)
                if repeats.size == 1:
                    repeats = int(repeats[0])
                    size = repeats * a.shape[axis or 0]
                elif a.shape[ax] == 1:
                    size = repeats = int(repeats.sum())
                else:
                    size = int(repeats.sum())
            else:
                size = np.nan
            if not isinstance(repeats, Integral):
                if repeats.ndim != 1:
                    raise ValueError("repeats should be 1-d tensor")
                broadcast_shape(repeats.shape, a.shape[ax : ax + 1])
        else:
            size = a.shape[axis or 0] * repeats

        shape = a.shape[:ax] + (size,) + a.shape[ax + 1 :]
        self.dtype = a.dtype
        self.sparse = a.issparse()

        inputs = [a]
        if isinstance(repeats, Tensor):
            inputs.append(repeats)
        else:
            self._repeats = repeats

        return self.new_tensor(inputs, shape, order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        a = op.input
        repeats = op.repeats
        axis = op.axis
        ax = axis or 0
        out = op.outputs[0]

        if has_unknown_shape(*op.inputs):
            yield

        if isinstance(repeats, TENSOR_TYPE):
            a, repeats = yield from unify_chunks(a, (repeats, (ax,)))

        nsplit = a.nsplits[axis or 0]

        if isinstance(repeats, Integral):
            new_nsplit = []
            for split in nsplit:
                s = max(split // repeats, 1)
                c = split // s
                new_nsplit.extend([s] * c)
                if split % s != 0:
                    new_nsplit.append(split % s)

            a = yield from recursive_tile(a.rechunk({ax: new_nsplit}))

        out_chunks = []
        ax_cum_count = np.cumsum((0,) + a.nsplits[ax])
        is_repeats_ndarray = isinstance(repeats, np.ndarray)
        for out_idx in itertools.product(*[range(len(s)) for s in a.nsplits]):
            in_chunk = a.cix[out_idx]
            ax_idx = out_idx[ax]
            if is_repeats_ndarray:
                start = ax_cum_count[ax_idx]
                stop = ax_cum_count[ax_idx + 1]
                rp = repeats[start:stop]
                size = int(rp.sum())
            elif not isinstance(repeats, Integral):
                rp = repeats.cix[
                    ax_idx,
                ]
                size = np.nan
            else:
                rp = repeats
                size = in_chunk.shape[ax] * rp

            chunk_inputs = [in_chunk]
            if isinstance(rp, TENSOR_CHUNK_TYPE):
                chunk_inputs.append(rp)

            chunk_shape = in_chunk.shape[:ax] + (size,) + in_chunk.shape[ax + 1 :]
            chunk_op = op.copy().reset_key()
            if len(chunk_inputs) < 2:
                # repeats is not chunk
                chunk_op._repeats = rp
            out_chunk = chunk_op.new_chunk(
                chunk_inputs, shape=chunk_shape, index=out_idx, order=out.order
            )
            out_chunks.append(out_chunk)

        nsplits = [
            tuple(
                c.shape[i]
                for c in out_chunks
                if all(idx == 0 for j, idx in enumerate(c.index) if j != i)
            )
            for i in range(len(out_chunks[0].shape))
        ]
        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, out.shape, order=out.order, chunks=out_chunks, nsplits=nsplits
        )

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        a = inputs[0]
        if len(inputs) > 1:
            repeats = inputs[1]
        else:
            repeats = op.repeats

        with device(device_id):
            ctx[op.outputs[0].key] = xp.repeat(a, repeats=repeats, axis=op.axis)


def repeat(a, repeats, axis=None):
    """
    Repeat elements of a tensor.

    Parameters
    ----------
    a : array_like
        Input tensor.
    repeats : int or tensor of ints
        The number of repetitions for each element.  `repeats` is broadcasted
        to fit the shape of the given axis.
    axis : int, optional
        The axis along which to repeat values.  By default, use the
        flattened input tensor, and return a flat output tensor.

    Returns
    -------
    repeated_tensor : Tensor
        Output array which has the same shape as `a`, except along
        the given axis.

    See Also
    --------
    tile : Tile a tensor.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.repeat(3, 4).execute()
    array([3, 3, 3, 3])
    >>> x = mt.array([[1,2],[3,4]])
    >>> mt.repeat(x, 2).execute()
    array([1, 1, 2, 2, 3, 3, 4, 4])
    >>> mt.repeat(x, 3, axis=1).execute()
    array([[1, 1, 1, 2, 2, 2],
           [3, 3, 3, 4, 4, 4]])
    >>> mt.repeat(x, [1, 2], axis=0).execute()
    array([[1, 2],
           [3, 4],
           [3, 4]])

    """
    op = TensorRepeat(axis=axis)
    return op(a, repeats)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from dataclasses import dataclass
from typing import Dict, List

from ... import oscar as mo
from ...lib.aio import alru_cache
from ...storage import StorageLevel
from ...utils import dataslots
from .core import DataManagerActor, WrappedStorageFileObject
from .handler import StorageHandlerActor

DEFAULT_TRANSFER_BLOCK_SIZE = 4 * 1024**2


logger = logging.getLogger(__name__)


class SenderManagerActor(mo.StatelessActor):
    def __init__(
        self,
        band_name: str = "numa-0",
        transfer_block_size: int = None,
        data_manager_ref: mo.ActorRefType[DataManagerActor] = None,
        storage_handler_ref: mo.ActorRefType[StorageHandlerActor] = None,
    ):
        self._band_name = band_name
        self._data_manager_ref = data_manager_ref
        self._storage_handler = storage_handler_ref
        self._transfer_block_size = transfer_block_size or DEFAULT_TRANSFER_BLOCK_SIZE

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"sender_manager_{band_name}"

    async def __post_create__(self):
        if self._storage_handler is None:  # for test
            self._storage_handler = await mo.actor_ref(
                self.address, StorageHandlerActor.gen_uid("numa-0")
            )

    @staticmethod
    @alru_cache
    async def get_receiver_ref(address: str, band_name: str):
        return await mo.actor_ref(
            address=address, uid=ReceiverManagerActor.gen_uid(band_name)
        )

    async def _send_data(
        self,
        receiver_ref: mo.ActorRefType["ReceiverManagerActor"],
        session_id: str,
        data_keys: List[str],
        level: StorageLevel,
        block_size: int,
    ):
        class BufferedSender:
            def __init__(self):
                self._buffers = []
                self._send_keys = []
                self._eof_marks = []

            async def flush(self):
                if self._buffers:
                    await receiver_ref.receive_part_data(
                        self._buffers, session_id, self._send_keys, self._eof_marks
                    )

                self._buffers = []
                self._send_keys = []
                self._eof_marks = []

            async def send(self, buffer, eof_mark, key):
                self._eof_marks.append(eof_mark)
                self._buffers.append(buffer)
                self._send_keys.append(key)
                if sum(len(b) for b in self._buffers) >= block_size:
                    await self.flush()

        sender = BufferedSender()
        open_reader_tasks = []
        for data_key in data_keys:
            open_reader_tasks.append(
                self._storage_handler.open_reader.delay(session_id, data_key)
            )
        readers = await self._storage_handler.open_reader.batch(*open_reader_tasks)

        for data_key, reader in zip(data_keys, readers):
            while True:
                part_data = await reader.read(block_size)
                # Notes on [How to decide whether the reader reaches EOF?]
                #
                # In some storage backend, e.g., the reported memory usage (i.e., the
                # `store_size`) may not same with the byte size that need to be transferred
                # when moving to a remote worker. Thus, we think the reader reaches EOF
                # when a `read` request returns nothing, rather than comparing the `sent_size`
                # and the `store_size`.
                #
                is_eof = not part_data  # can be non-empty bytes, empty bytes and None
                await sender.send(part_data, is_eof, data_key)
                if is_eof:
                    break
        await sender.flush()

    @mo.extensible
    async def send_batch_data(
        self,
        session_id: str,
        data_keys: List[str],
        address: str,
        level: StorageLevel,
        band_name: str = "numa-0",
        block_size: int = None,
        error: str = "raise",
    ):
        logger.debug(
            "Begin to send data (%s, %s) to %s", session_id, data_keys, address
        )
        block_size = block_size or self._transfer_block_size
        receiver_ref: mo.ActorRefType[
            ReceiverManagerActor
        ] = await self.get_receiver_ref(address, band_name)
        get_infos = []
        pin_tasks = []
        for data_key in data_keys:
            get_infos.append(
                self._data_manager_ref.get_data_info.delay(
                    session_id, data_key, self._band_name, error
                )
            )
            pin_tasks.append(
                self._data_manager_ref.pin.delay(
                    session_id, data_key, self._band_name, error
                )
            )
        await self._data_manager_ref.pin.batch(*pin_tasks)
        infos = await self._data_manager_ref.get_data_info.batch(*get_infos)
        filtered = [
            (data_info, data_key)
            for data_info, data_key in zip(infos, data_keys)
            if data_info is not None
        ]
        if filtered:
            infos, data_keys = zip(*filtered)
        else:  # pragma: no cover
            # no data to be transferred
            return
        data_sizes = [info.store_size for info in infos]
        if level is None:
            level = infos[0].level
        is_transferring_list = await receiver_ref.open_writers(
            session_id, data_keys, data_sizes, level
        )
        to_send_keys = []
        to_wait_keys = []
        for data_key, is_transferring in zip(data_keys, is_transferring_list):
            if is_transferring:
                to_wait_keys.append(data_key)
            else:
                to_send_keys.append(data_key)

        if to_send_keys:
            await self._send_data(
                receiver_ref, session_id, to_send_keys, level, block_size
            )
        if to_wait_keys:
            await receiver_ref.wait_transfer_done(session_id, to_wait_keys)
        unpin_tasks = []
        for data_key in data_keys:
            unpin_tasks.append(
                self._data_manager_ref.unpin.delay(
                    session_id, [data_key], self._band_name, error="ignore"
                )
            )
        await self._data_manager_ref.unpin.batch(*unpin_tasks)
        logger.debug(
            "Finish sending data (%s, %s) to %s, total size is %s",
            session_id,
            data_keys,
            address,
            sum(data_sizes),
        )


@dataslots
@dataclass
class WritingInfo:
    writer: WrappedStorageFileObject
    size: int
    level: StorageLevel
    event: asyncio.Event
    ref_counts: int


class ReceiverManagerActor(mo.StatelessActor):
    def __init__(
        self,
        quota_refs: Dict,
        storage_handler_ref: mo.ActorRefType[StorageHandlerActor] = None,
    ):
        self._quota_refs = quota_refs
        self._storage_handler = storage_handler_ref
        self._writing_infos: Dict[tuple, WritingInfo] = dict()
        self._lock = asyncio.Lock()

    async def __post_create__(self):
        if self._storage_handler is None:  # for test
            self._storage_handler = await mo.actor_ref(
                self.address, StorageHandlerActor.gen_uid("numa-0")
            )

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"receiver_manager_{band_name}"

    def _decref_writing_key(self, session_id: str, data_key: str):
        self._writing_infos[(session_id, data_key)].ref_counts -= 1
        if self._writing_infos[(session_id, data_key)].ref_counts == 0:
            del self._writing_infos[(session_id, data_key)]

    async def create_writers(
        self,
        session_id: str,
        data_keys: List[str],
        data_sizes: List[int],
        level: StorageLevel,
    ):
        tasks = dict()
        data_key_to_size = dict()
        being_processed = []
        for data_key, data_size in zip(data_keys, data_sizes):
            data_key_to_size[data_key] = data_size
            if (session_id, data_key) not in self._writing_infos:
                being_processed.append(False)
                tasks[data_key] = self._storage_handler.open_writer.delay(
                    session_id, data_key, data_size, level, request_quota=False
                )
            else:
                being_processed.append(True)
                self._writing_infos[(session_id, data_key)].ref_counts += 1
        if tasks:
            writers = await self._storage_handler.open_writer.batch(
                *tuple(tasks.values())
            )
            for data_key, writer in zip(tasks, writers):
                self._writing_infos[(session_id, data_key)] = WritingInfo(
                    writer, data_key_to_size[data_key], level, asyncio.Event(), 1
                )
        return being_processed

    async def open_writers(
        self,
        session_id: str,
        data_keys: List[str],
        data_sizes: List[int],
        level: StorageLevel,
    ):
        async with self._lock:
            await self._storage_handler.request_quota_with_spill(level, sum(data_sizes))
            future = asyncio.create_task(
                self.create_writers(session_id, data_keys, data_sizes, level)
            )
            try:
                return await future
            except asyncio.CancelledError:
                await self._quota_refs[level].release_quota(sum(data_sizes))
                future.cancel()
                raise

    async def do_write(
        self, data: list, session_id: str, data_keys: List[str], eof_marks: List[bool]
    ):
        # close may be a high-cost operation, use create_task
        close_tasks = []
        finished_keys = []
        for data, data_key, is_eof in zip(data, data_keys, eof_marks):
            writer = self._writing_infos[(session_id, data_key)].writer
            if data:
                await writer.write(data)
            if is_eof:
                close_tasks.append(writer.close())
                finished_keys.append(data_key)
        await asyncio.gather(*close_tasks)
        async with self._lock:
            for data_key in finished_keys:
                event = self._writing_infos[(session_id, data_key)].event
                event.set()
                self._decref_writing_key(session_id, data_key)

    async def receive_part_data(
        self, data: list, session_id: str, data_keys: List[str], eof_marks: List[bool]
    ):
        write_task = asyncio.create_task(
            self.do_write(data, session_id, data_keys, eof_marks)
        )
        try:
            await asyncio.shield(write_task)
        except asyncio.CancelledError:
            async with self._lock:
                for data_key in data_keys:
                    if (session_id, data_key) in self._writing_infos:
                        if self._writing_infos[(session_id, data_key)].ref_counts == 1:
                            info = self._writing_infos[(session_id, data_key)]
                            await self._quota_refs[info.level].release_quota(info.size)
                            await self._storage_handler.delete(
                                session_id, data_key, error="ignore"
                            )
                            await info.writer.clean_up()
                            info.event.set()
                            self._decref_writing_key(session_id, data_key)
                            write_task.cancel()
                            await write_task
            raise

    async def wait_transfer_done(self, session_id, data_keys):
        await asyncio.gather(
            *[self._writing_infos[(session_id, key)].event.wait() for key in data_keys]
        )
        async with self._lock:
            for data_key in data_keys:
                self._decref_writing_key(session_id, data_key)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Filename globbing utility, modified from python glob.

obviously，this implementation is not optimal, it will cause too many
oss requests. Lately, We can then convert the glob expression into
a regular expression, and then match the oss key list.
But before that, we need to figure out how to deal with magic char
in oss key, such like oss glob: oss://bucket/[key]/*, the key
oss://bucket/[key]/a exactly exists.

Notes:
    OSS need a bucket to specify the file or dir, the "**" patten is
    not supported. So _isrecursive(pattern) is removed.
"""

import fnmatch
import os
import re

from .common import oss_exists, oss_isdir, oss_scandir

__all__ = ["glob", "iglob", "escape"]


def glob(pathname, *, recursive=False):
    """Return a list of paths matching a pathname pattern.
    The pattern may contain simple shell-style wildcards a la
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.
    """
    return list(iglob(pathname, recursive=recursive))


def iglob(pathname, *, recursive=False):
    """Return an iterator which yields the paths matching a pathname pattern.
    The pattern may contain simple shell-style wildcards like
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.
    """
    it = _iglob(pathname, recursive, False)
    return it


def _iglob(pathname, recursive, dironly):
    dirname, basename = os.path.split(pathname)
    if not has_magic(pathname):
        assert not dironly
        if basename:
            if oss_exists(pathname):
                yield pathname
        else:
            # Patterns ending with a slash should match only directories
            if oss_isdir(dirname):
                yield pathname
        return
    # dirname will not be None in oss path.
    #  Prevent an infinite recursion if a drive or UNC path
    # contains magic characters (i.e. r'\\?\C:').
    if dirname != pathname and has_magic(dirname):
        dirs = _iglob(dirname, recursive, True)
    else:
        dirs = [dirname]
    if has_magic(basename):
        glob_in_dir = _glob1
    else:
        glob_in_dir = _glob0
    for dirname in dirs:
        for name in glob_in_dir(dirname, basename, dironly):
            yield os.path.join(dirname, name)


# These 2 helper functions non-recursively glob inside a literal directory.
# They return a list of basenames.  _glob1 accepts a pattern while _glob0
# takes a literal basename (so it only has to check for its existence).


def _glob1(dirname, pattern, dironly):
    names = list(_iterdir(dirname, dironly))
    if not _ishidden(pattern):
        names = (x for x in names if not _ishidden(x))
    return fnmatch.filter(names, pattern)


def _glob0(dirname, basename, dironly):
    if not basename:
        # `os.path.split()` returns an empty basename for paths ending with a
        # directory separator.  'q*x/' should match only directories.
        if oss_isdir(dirname):
            return [basename]
    else:
        if oss_exists(os.path.join(dirname, basename)):
            return [basename]
    return []


# If dironly is false, yields all file names inside a directory.
# If dironly is true, yields only directory names.
# An oss path must contain a dirname.
def _iterdir(dirname, dironly):
    for entry in oss_scandir(dirname):
        if not dironly or entry.is_dir():
            yield entry.name
    return


magic_check = re.compile("([*?[])")
magic_check_bytes = re.compile(b"([*?[])")


def has_magic(s):
    if isinstance(s, bytes):
        match = magic_check_bytes.search(s)
    else:
        match = magic_check.search(s)
    return match is not None


def _ishidden(path):
    return False


def escape(pathname):
    """Escape all special characters."""
    # Escaping is done by wrapping any of "*?[" between square brackets.
    # Metacharacters do not work in the drive part and shouldn't be escaped.
    drive, pathname = os.path.splitdrive(pathname)
    if isinstance(pathname, bytes):
        pathname = magic_check_bytes.sub(rb"[\1]", pathname)
    else:
        pathname = magic_check.sub(r"[\1]", pathname)
    return drive + pathname



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...serialization.serializables import FieldTypes, StringField, ListField
from .base import Operand
from .core import TileableOperandMixin


class Fetch(Operand):
    _op_type_ = opcodes.FETCH

    source_key = StringField("source_key", default=None)


class FetchMixin(TileableOperandMixin):
    def check_inputs(self, inputs):
        # no inputs
        if inputs and len(inputs) > 0:
            raise ValueError(f"{type(self).__name__} has no inputs")

    @classmethod
    def tile(cls, op):
        raise NotImplementedError("Fetch tile cannot be handled by operand itself")

    @classmethod
    def execute(cls, ctx, op):
        """
        Fetch operand needs nothing to do.
        """


class FetchShuffle(Operand):
    _op_type_ = opcodes.FETCH_SHUFFLE

    source_keys = ListField("source_keys", FieldTypes.string)
    source_idxes = ListField("source_idxes", FieldTypes.tuple(FieldTypes.uint64))
    source_mappers = ListField("source_mappers", FieldTypes.uint16)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import List, Dict, Set

from ....resource import Resource
from ....typing import BandType
from ...core import NodeRole
from ..core import NodeStatus


class AbstractClusterAPI:
    @staticmethod
    def _calc_statuses(
        statuses: Set[NodeStatus] = None, exclude_statuses: Set[NodeStatus] = None
    ) -> Set[NodeStatus]:
        if statuses:
            return statuses
        elif exclude_statuses is not None:
            return set(NodeStatus.__members__.values()).difference(exclude_statuses)
        else:
            return {NodeStatus.READY}

    @abstractmethod
    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        """
        Get supervisor addresses

        Returns
        -------
        out
            list of supervisors
        """

    @abstractmethod
    async def watch_supervisors(self):
        """
        Watch supervisor addresses

        Returns
        -------
        out
            generator of list of supervisors
        """

    @abstractmethod
    async def watch_nodes(
        self,
        role: NodeRole,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> List[Dict[str, Dict]]:
        """
        Watch changes of workers

        Returns
        -------
        out: List[Dict[str, Dict]]
            dict of worker resources by addresses and bands
        """

    @abstractmethod
    async def get_nodes_info(
        self,
        nodes: List[str] = None,
        role: NodeRole = None,
        env: bool = False,
        resource: bool = False,
        detail: bool = False,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        """
        Get worker info

        Parameters
        ----------
        nodes
            address of nodes
        role
            roles of nodes
        env
            receive env info
        resource
            receive resource info
        detail
            receive detail info

        Returns
        -------
        out: Dict
            info of worker
        """

    @abstractmethod
    async def get_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ) -> Dict[BandType, Resource]:
        """
        Get all bands that can be used for computation.

        Returns
        -------
        band_to_resource : dict
            Band to resource.
        """

    @abstractmethod
    async def watch_all_bands(
        self,
        role: NodeRole = None,
        statuses: Set[NodeStatus] = None,
        exclude_statuses: Set[NodeStatus] = None,
    ):
        """
        Watch all bands that can be used for computation.

        Returns
        -------
        band_to_resource : dict
            Band to resource.
        """

    @abstractmethod
    async def get_mars_versions(self) -> List[str]:
        """
        Get versions used in current Mars cluster

        Returns
        -------
        version_list : list
            List of versions
        """

    @abstractmethod
    async def get_node_pool_configs(self, address: str) -> List[Dict]:
        """
        Get pool configs of a Mars node

        Returns
        -------
        config_list : List[Dict]
            List of configs for all pool processes
        """

    async def get_node_thread_stacks(self, address: str) -> List[Dict[int, List[str]]]:
        """
        Get current thread pool stacks of a Mars node

        Parameters
        ----------
        address

        Returns
        -------

        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorComplexFFTMixin, validate_fft, TensorStandardFFT


class TensorFFT(TensorStandardFFT, TensorComplexFFTMixin):
    _op_type_ = OperandDef.FFT

    def __init__(self, n=None, axis=-1, norm=None, **kw):
        super().__init__(_n=n, _axis=axis, _norm=norm, **kw)


def fft(a, n=None, axis=-1, norm=None):
    """
    Compute the one-dimensional discrete Fourier Transform.

    This function computes the one-dimensional *n*-point discrete Fourier
    Transform (DFT) with the efficient Fast Fourier Transform (FFT)
    algorithm [CT].

    Parameters
    ----------
    a : array_like
        Input tensor, can be complex.
    n : int, optional
        Length of the transformed axis of the output.
        If `n` is smaller than the length of the input, the input is cropped.
        If it is larger, the input is padded with zeros.  If `n` is not given,
        the length of the input along the axis specified by `axis` is used.
    axis : int, optional
        Axis over which to compute the FFT.  If not given, the last axis is
        used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.

    Raises
    ------
    IndexError
        if `axes` is larger than the last axis of `a`.

    See Also
    --------
    mt.fft : for definition of the DFT and conventions used.
    ifft : The inverse of `fft`.
    fft2 : The two-dimensional FFT.
    fftn : The *n*-dimensional FFT.
    rfftn : The *n*-dimensional FFT of real input.
    fftfreq : Frequency bins for given FFT parameters.

    Notes
    -----
    FFT (Fast Fourier Transform) refers to a way the discrete Fourier
    Transform (DFT) can be calculated efficiently, by using symmetries in the
    calculated terms.  The symmetry is highest when `n` is a power of 2, and
    the transform is therefore most efficient for these sizes.

    The DFT is defined, with the conventions used in this implementation, in
    the documentation for the `numpy.fft` module.

    References
    ----------
    .. [CT] Cooley, James W., and John W. Tukey, 1965, "An algorithm for the
            machine calculation of complex Fourier series," *Math. Comput.*
            19: 297-301.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fft.fft(mt.exp(2j * mt.pi * mt.arange(8) / 8)).execute()
    array([-2.33486982e-16+1.14423775e-17j,  8.00000000e+00-6.89018570e-16j,
            2.33486982e-16+2.33486982e-16j,  0.00000000e+00+0.00000000e+00j,
           -1.14423775e-17+2.33486982e-16j,  0.00000000e+00+1.99159850e-16j,
            1.14423775e-17+1.14423775e-17j,  0.00000000e+00+0.00000000e+00j])

    In this example, real input has an FFT which is Hermitian, i.e., symmetric
    in the real part and anti-symmetric in the imaginary part, as described in
    the `numpy.fft` documentation:

    >>> import matplotlib.pyplot as plt
    >>> t = mt.arange(256)
    >>> sp = mt.fft.fft(mt.sin(t))
    >>> freq = mt.fft.fftfreq(t.shape[-1])
    >>> plt.plot(freq.execute(), sp.real.execute(), freq.execute(), sp.imag.execute())
    [<matplotlib.lines.Line2D object at 0x...>, <matplotlib.lines.Line2D object at 0x...>]
    >>> plt.show()

    """
    a = astensor(a)
    validate_fft(a, axis, norm)
    op = TensorFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import builtins
import inspect
import itertools
from operator import attrgetter
from typing import List, Callable, Generator
from weakref import WeakSet, WeakKeyDictionary

import numpy as np

from ...serialization.serializables import FieldTypes, BoolField, TupleField
from ...typing import OperandType, TileableType, ChunkType
from ...utils import on_serialize_shape, on_deserialize_shape, on_serialize_nsplits
from ..base import Base
from ..mode import enter_mode
from .chunks import Chunk
from .core import EntityData, Entity
from .executable import _ExecutableMixin


class NotSupportTile(Exception):
    pass


class OperandTilesHandler:
    _handlers = dict()

    @classmethod
    def _get_op_cls(cls, op: OperandType):
        if isinstance(op, type):
            return op
        return type(op)

    @classmethod
    def register(
        cls, op: OperandType, tile_handler: Callable[[OperandType], TileableType]
    ):
        cls._handlers[cls._get_op_cls(op)] = tile_handler

    @classmethod
    def unregister(cls, op: OperandType):
        del cls._handlers[cls._get_op_cls(op)]

    @classmethod
    def get_handler(
        cls, op: OperandType
    ) -> Callable[[OperandType], List[TileableType]]:
        op_cls = cls._get_op_cls(op)
        return cls._handlers.get(op_cls, op_cls.tile)

    @classmethod
    def tile(
        cls, tileables: List[TileableType]
    ) -> Generator[List[ChunkType], List[ChunkType], List[TileableType]]:
        op = tileables[0].op
        # pre tile
        op.pre_tile(op)
        tiled_result = None
        try:
            tile_handler = cls.get_handler(op)
            if inspect.isgeneratorfunction(tile_handler):
                # op.tile can be a generator function,
                # each time an operand yield some chunks,
                # they will be put into ChunkGraph and executed first.
                # After execution, resume from the yield place.
                tiled_result = yield from tile_handler(op)
            else:
                # without iterative tiling
                tiled_result = tile_handler(op)
        finally:
            op.post_tile(op, tiled_result)

        if not isinstance(tiled_result, list):
            tiled_result = [tiled_result]
        tiled_results = [t.data if hasattr(t, "data") else t for t in tiled_result]
        assert len(tileables) == len(tiled_results)
        if any(inspect.isgenerator(r) for r in tiled_results):  # pragma: no cover
            raise TypeError(f"tiled result cannot be generator when tiling {op}")
        cls._assign_to(tiled_results, tileables)
        return tileables

    @classmethod
    def _assign_to(
        cls,
        tile_after_tensor_datas: List["TileableData"],
        tile_before_tensor_datas: List["TileableData"],
    ):
        assert len(tile_after_tensor_datas) == len(tile_before_tensor_datas)

        for tile_after_tensor_data, tile_before_tensor_data in zip(
            tile_after_tensor_datas, tile_before_tensor_datas
        ):
            if tile_before_tensor_data is None:
                # garbage collected
                continue
            tile_after_tensor_data.copy_to(tile_before_tensor_data)
            tile_before_tensor_data.op.outputs = tile_before_tensor_datas

    @enter_mode(kernel=True)
    def dispatch(self, op: OperandType):
        op_cls = self._get_op_cls(op)
        tiled = None
        cause = None

        if op_cls in self._handlers:
            tiled = self._handlers[op_cls](op)
        else:
            try:
                tiled = op_cls.tile(op)
            except NotImplementedError as ex:
                cause = ex
                for super_cls in op_cls.__mro__:
                    if super_cls in self._handlers:
                        h = self._handlers[op_cls] = self._handlers[super_cls]
                        tiled = h(op)
                        break

        if tiled is not None:
            return tiled if isinstance(tiled, list) else [tiled]
        else:
            raise NotImplementedError(f"{type(op)} does not support tile") from cause


handler = OperandTilesHandler()
register = OperandTilesHandler.register
unregister = OperandTilesHandler.unregister


class _ChunksIndexer:
    __slots__ = ("_tileable",)

    def __init__(self, tileable):
        self._tileable = tileable

    def __getitem__(self, item):
        """
        The indices for `cix` can be [x, y] or [x, :].
        For the former the result will be a single chunk,
        and for the later the result will be a list of chunks (flattened).

        The length of indices must be the same with `chunk_shape` of tileable.
        """
        if isinstance(item, int):
            item = (item,)
        if isinstance(item, tuple):
            if len(item) == 0 and self._tileable.is_scalar():
                return self._tileable.chunks[0]
            if len(item) != self._tileable.ndim:
                raise ValueError(
                    f"Cannot get chunk by {item}, "
                    f"expect length {self._tileable.ndim}"
                )
            slices, singleton = [], True
            for it, dim in zip(item, self._tileable.chunk_shape):
                if isinstance(it, slice):
                    slices.append(range(dim)[it])
                    singleton = False
                elif np.issubdtype(type(it), np.integer):
                    slices.append([it if it >= 0 else dim + it])
                else:
                    raise TypeError(
                        f"Cannot get chunk by {it}, "
                        f"invalid value has type {type(it)}"
                    )

            indexes = tuple(zip(*itertools.product(*slices)))

            flat_index = np.ravel_multi_index(indexes, self._tileable.chunk_shape)
            if singleton:
                return self._tileable._chunks[flat_index[0]]
            else:
                return [self._tileable._chunks[idx] for idx in flat_index]

        raise ValueError(f"Cannot get {type(self._tileable).__name__} chunk by {item}")


class EntityDataModificationHandler:
    def __init__(self):
        self._data_to_entities = WeakKeyDictionary()

    def _add_observer(self, data, entity):
        # only tileable data should be considered
        assert isinstance(data, TileableData)
        assert isinstance(entity, Tileable)

        if data not in self._data_to_entities:
            self._data_to_entities[data] = WeakSet()

        self._data_to_entities[data].add(entity)

    @enter_mode(build=True)
    def add_observer(self, data, entity):
        self._add_observer(data, entity)

    def _update_observe_data(self, observer, data, new_data):
        self._data_to_entities.get(data, set()).discard(observer)
        self._add_observer(new_data, observer)

    @staticmethod
    def _set_data(entity, data):
        entity._data.detach(entity)
        entity._data = data
        data.attach(entity)

    @staticmethod
    def _get_data(obj):
        return obj.data if isinstance(obj, Entity) else obj

    @enter_mode(build=True)
    def data_changed(self, old_data, new_data):
        notified = set()
        processed_data = set()
        old_to_new = {old_data: new_data}
        q = [old_data]
        while len(q) > 0:
            data = q.pop()

            # handle entities
            for entity in data.entities:
                self._set_data(entity, old_to_new[data])
                notified.add(entity)

            observers = {
                ob
                for ob in self._data_to_entities.pop(data, set())
                if ob not in notified
            }
            for ob in observers:
                new_data = self._get_data(ob.op.on_input_modify(old_to_new[data]))
                old_data = ob.data
                self._update_observe_data(ob, ob.data, new_data)
                old_to_new[old_data] = new_data
                if old_data not in processed_data:
                    q.append(old_data)
                    processed_data.add(old_data)
                notified.add(ob)

            if data.op.create_view:
                old_input_data = data.inputs[0]
                new_input_data = self._get_data(
                    data.op.on_output_modify(old_to_new[data])
                )
                old_to_new[old_input_data] = new_input_data
                if old_input_data not in processed_data:
                    q.append(old_input_data)
                    processed_data.add(old_input_data)


entity_view_handler = EntityDataModificationHandler()


class TileableData(EntityData, _ExecutableMixin):
    __slots__ = "_cix", "_entities", "_executed_sessions"
    _no_copy_attrs_ = Base._no_copy_attrs_ | {"_cix"}

    # optional fields
    # `nsplits` means the sizes of chunks for each dimension
    _nsplits = TupleField(
        "nsplits",
        FieldTypes.tuple(FieldTypes.uint64),
        on_serialize=on_serialize_nsplits,
    )
    # cache tileable data, if true, this data will be materialized
    cache = BoolField("cache", default=False)

    def __init__(self: TileableType, *args, **kwargs):
        if kwargs.get("_nsplits", None) is not None:
            kwargs["_nsplits"] = tuple(tuple(s) for s in kwargs["_nsplits"])

        super().__init__(*args, **kwargs)

        try:
            chunks = self._chunks
            if chunks:
                self._chunks = sorted(chunks, key=attrgetter("index"))
        except AttributeError:  # pragma: no cover
            pass

        self._entities = WeakSet()
        self._executed_sessions = []

    @property
    def chunk_shape(self):
        if hasattr(self, "_nsplits") and self._nsplits is not None:
            return tuple(map(len, self._nsplits))

    @property
    def chunks(self) -> List[Chunk]:
        return getattr(self, "_chunks", None)

    @property
    def nsplits(self):
        return getattr(self, "_nsplits", None)

    @nsplits.setter
    def nsplits(self, new_nsplits):
        self._nsplits = new_nsplits

    @property
    def params(self) -> dict:
        # params return the properties which useful to rebuild a new tileable object
        return dict()

    @property
    def cix(self):
        if self.ndim == 0:
            return _ChunksIndexer(self)

        try:
            if getattr(self, "_cix", None) is None:
                self._cix = _ChunksIndexer(self)
            return self._cix
        except (TypeError, ValueError):
            return _ChunksIndexer(self)

    @property
    def entities(self):
        return self._entities

    def is_coarse(self):
        if not hasattr(self, "_chunks"):
            return True
        if not self._chunks:
            return True
        return False

    def attach(self, entity):
        self._entities.add(entity)

    def detach(self, entity):
        self._entities.discard(entity)


class Tileable(Entity):
    def __init__(self, data: TileableType = None, **kw):
        super().__init__(data=data, **kw)
        data = self._data
        if data is not None:
            data.attach(self)
            if data.op.create_view:
                entity_view_handler.add_observer(data.inputs[0], self)

    def __copy__(self):
        return self._view()

    def _view(self):
        return super().copy()

    def copy(self: TileableType) -> TileableType:
        new_op = self.op.copy()
        if new_op.create_view:
            # if the operand is a view, make it a copy
            new_op.create_view = False
        params = []
        for o in self.op.outputs:
            param = o.params
            param["_key"] = o.key
            param.update(o.extra_params)
            params.append(param)
        new_outs = new_op.new_tileables(
            self.op.inputs, kws=params, output_limit=len(params)
        )
        pos = -1
        for i, out in enumerate(self.op.outputs):
            # create a ref to copied one
            new_out = new_outs[i]
            if not hasattr(new_out.data, "_siblings"):
                new_out.data._siblings = []
            new_out.data._siblings.append(out)

            if self._data is out:
                pos = i
                break
        assert pos >= 0
        return new_outs[pos]

    @Entity.data.setter
    def data(self, new_data):
        self._check_data(new_data)
        if self._data is None:
            self._data = new_data
            self._data.attach(self)
        else:
            entity_view_handler.data_changed(self._data, new_data)


TILEABLE_TYPE = (Tileable, TileableData)


class HasShapeTileableData(TileableData):
    # required fields
    _shape = TupleField(
        "shape",
        FieldTypes.int64,
        on_serialize=on_serialize_shape,
        on_deserialize=on_deserialize_shape,
    )

    @property
    def ndim(self):
        return len(self.shape)

    def __len__(self):
        try:
            return int(self.shape[0])
        except (IndexError, ValueError):  # pragma: no cover
            return 0

    @property
    def shape(self):
        if hasattr(self, "_shape") and self._shape is not None:
            return self._shape
        if hasattr(self, "_nsplits") and self._nsplits is not None:
            self._shape = tuple(builtins.sum(nsplit) for nsplit in self._nsplits)
            return self._shape

    def _update_shape(self, new_shape):
        self._shape = new_shape

    @property
    def size(self):
        return np.prod(self.shape).item()

    @property
    def params(self):
        # params return the properties which useful to rebuild a new tileable object
        return {"shape": self.shape}

    def _equals(self, o):
        return self is o


class HasShapeTileable(Tileable):
    __slots__ = ()

    @property
    def shape(self):
        return self._data.shape

    @property
    def ndim(self):
        return self._data.ndim

    @property
    def size(self):
        return self._data.size

    def execute(self, session=None, **kw):
        result = self.data.execute(session=session, **kw)
        if isinstance(result, TILEABLE_TYPE):
            return self
        else:
            return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import CustomReduction
from .aggregation import DataFrameAggregate

from .sum import DataFrameSum
from .prod import DataFrameProd
from .max import DataFrameMax
from .min import DataFrameMin
from .count import DataFrameCount
from .mean import DataFrameMean
from .var import DataFrameVar
from .all import DataFrameAll
from .any import DataFrameAny
from .skew import DataFrameSkew
from .kurtosis import DataFrameKurtosis
from .sem import DataFrameSem
from .reduction_size import DataFrameSize
from .str_concat import DataFrameStrConcat, build_str_concat_object
from .custom_reduction import DataFrameCustomReduction

from .cummax import DataFrameCummax
from .cummin import DataFrameCummin
from .cumprod import DataFrameCumprod
from .cumsum import DataFrameCumsum

from .nunique import DataFrameNunique
from .unique import DataFrameUnique, unique


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE
    from .aggregation import aggregate
    from .sum import sum_series, sum_dataframe
    from .prod import prod_series, prod_dataframe
    from .max import max_series, max_dataframe, max_index
    from .min import min_series, min_dataframe, min_index
    from .count import count_series, count_dataframe
    from .mean import mean_series, mean_dataframe
    from .var import var_series, var_dataframe
    from .std import std_series, std_dataframe
    from .all import all_series, all_dataframe, all_index
    from .any import any_series, any_dataframe, any_index
    from .cummax import cummax
    from .cummin import cummin
    from .cumprod import cumprod
    from .cumsum import cumsum
    from .nunique import nunique_dataframe, nunique_series
    from .sem import sem_dataframe, sem_series
    from .skew import skew_dataframe, skew_series
    from .kurtosis import kurt_dataframe, kurt_series
    from .reduction_size import size_dataframe, size_series

    funcs = [
        ("sum", sum_series, sum_dataframe),
        ("prod", prod_series, prod_dataframe),
        ("product", prod_series, prod_dataframe),
        ("max", max_series, max_dataframe),
        ("min", min_series, min_dataframe),
        ("count", count_series, count_dataframe),
        ("mean", mean_series, mean_dataframe),
        ("var", var_series, var_dataframe),
        ("std", std_series, std_dataframe),
        ("all", all_series, all_dataframe),
        ("any", any_series, any_dataframe),
        ("cummax", cummax, cummax),
        ("cummin", cummin, cummin),
        ("cumprod", cumprod, cumprod),
        ("cumsum", cumsum, cumsum),
        ("agg", aggregate, aggregate),
        ("aggregate", aggregate, aggregate),
        ("nunique", nunique_series, nunique_dataframe),
        ("sem", sem_series, sem_dataframe),
        ("skew", skew_series, skew_dataframe),
        ("kurt", kurt_series, kurt_dataframe),
        ("kurtosis", kurt_series, kurt_dataframe),
        ("unique", unique, None),
        ("_reduction_size", size_dataframe, size_series),
    ]
    for func_name, series_func, df_func in funcs:
        if df_func is not None:  # pragma: no branch
            for t in DATAFRAME_TYPE:
                setattr(t, func_name, df_func)
        if series_func is not None:  # pragma: no branch
            for t in SERIES_TYPE:
                setattr(t, func_name, series_func)

    for t in INDEX_TYPE:
        setattr(t, "agg", aggregate)
        setattr(t, "aggregate", aggregate)
        setattr(t, "all", all_index)
        setattr(t, "any", any_index)
        setattr(t, "min", min_index)
        setattr(t, "max", max_index)


_install()
del _install



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...serialization.serializables import (
    KeyField,
    AnyField,
    StringField,
    DataTypeField,
    BoolField,
    Int32Field,
)
from ...tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ...tensor.datasource import (
    empty,
    tensor as astensor,
    from_series as tensor_from_series,
    from_dataframe as tensor_from_dataframe,
)
from ...tensor.statistics.quantile import quantile as tensor_quantile
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..core import DATAFRAME_TYPE
from ..datasource.from_tensor import series_from_tensor, dataframe_from_tensor
from ..initializer import DataFrame as create_df
from ..utils import parse_index, build_empty_df, find_common_type, validate_axis


class DataFrameQuantile(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.QUANTILE

    _input = KeyField("input")
    _q = AnyField("q")
    _axis = Int32Field("axis")
    _numeric_only = BoolField("numeric_only")
    _interpolation = StringField("interpolation")

    _dtype = DataTypeField("dtype")

    def __init__(
        self,
        q=None,
        interpolation=None,
        axis=None,
        numeric_only=None,
        dtype=None,
        gpu=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _q=q,
            _interpolation=interpolation,
            _axis=axis,
            _numeric_only=numeric_only,
            _dtype=dtype,
            _output_types=output_types,
            gpu=gpu,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def q(self):
        return self._q

    @property
    def interpolation(self):
        return self._interpolation

    @property
    def axis(self):
        return self._axis

    @property
    def numeric_only(self):
        return self._numeric_only

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if isinstance(self._q, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._q = self._inputs[-1]

    def _calc_dtype_on_axis_1(self, a, dtypes):
        quantile_dtypes = []
        for name in dtypes.index:
            dt = tensor_quantile(
                tensor_from_series(a[name]),
                self._q,
                interpolation=self._interpolation,
                handle_non_numeric=not self._numeric_only,
            ).dtype
            quantile_dtypes.append(dt)
        return find_common_type(quantile_dtypes)

    def _call_dataframe(self, a, inputs):
        if self._numeric_only:
            empty_df = build_empty_df(a.dtypes)
            dtypes = empty_df._get_numeric_data().dtypes
        else:
            dtypes = a.dtypes
        if isinstance(self._q, TENSOR_TYPE):
            q_val = self._q
            pd_index = pd.Index([], dtype=q_val.dtype)
            name = None
            store_index_value = False
        else:
            q_val = np.asanyarray(self._q)
            pd_index = pd.Index(q_val)
            name = self._q if q_val.size == 1 else None
            store_index_value = True
        tokenize_objects = (a, q_val, self._interpolation, type(self).__name__)

        if q_val.ndim == 0 and self._axis == 0:
            index_value = parse_index(dtypes.index, store_data=store_index_value)
            shape = (len(dtypes),)
            # calc dtype
            dtype = self._calc_dtype_on_axis_1(a, dtypes)
            return self.new_series(
                inputs,
                shape=shape,
                dtype=dtype,
                index_value=index_value,
                name=name or dtypes.index.name,
            )
        elif q_val.ndim == 0 and self._axis == 1:
            index_value = a.index_value
            shape = (len(a),)
            # calc dtype
            dt = tensor_quantile(
                empty(a.shape[1], dtype=find_common_type(list(dtypes))),
                self._q,
                interpolation=self._interpolation,
                handle_non_numeric=not self._numeric_only,
            ).dtype
            return self.new_series(
                inputs,
                shape=shape,
                dtype=dt,
                index_value=index_value,
                name=name or index_value.name,
            )
        elif q_val.ndim == 1 and self._axis == 0:
            shape = (len(q_val), len(dtypes))
            index_value = parse_index(
                pd_index, *tokenize_objects, store_data=store_index_value
            )
            dtype_list = []
            for name in dtypes.index:
                dtype_list.append(
                    tensor_quantile(
                        tensor_from_series(a[name]),
                        self._q,
                        interpolation=self._interpolation,
                        handle_non_numeric=not self._numeric_only,
                    ).dtype
                )
            dtypes = pd.Series(dtype_list, index=dtypes.index)
            return self.new_dataframe(
                inputs,
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=parse_index(dtypes.index, store_data=True),
            )
        else:
            assert q_val.ndim == 1 and self._axis == 1
            shape = (len(q_val), a.shape[0])
            index_value = parse_index(
                pd_index, *tokenize_objects, store_data=store_index_value
            )
            pd_columns = a.index_value.to_pandas()
            dtype_list = np.full(len(pd_columns), self._calc_dtype_on_axis_1(a, dtypes))
            dtypes = pd.Series(dtype_list, index=pd_columns)
            return self.new_dataframe(
                inputs,
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=parse_index(
                    dtypes.index, store_data=True, key=a.index_value.key
                ),
            )

    def _call_series(self, a, inputs):
        if isinstance(self._q, TENSOR_TYPE):
            q_val = self._q
            index_val = pd.Index([], dtype=q_val.dtype)
            store_index_value = False
        else:
            q_val = np.asanyarray(self._q)
            index_val = pd.Index(q_val)
            store_index_value = True

        # get dtype by tensor
        a_t = astensor(a)
        self._dtype = dtype = tensor_quantile(
            a_t,
            self._q,
            interpolation=self._interpolation,
            handle_non_numeric=not self._numeric_only,
        ).dtype

        if q_val.ndim == 0:
            return self.new_scalar(inputs, dtype=dtype)
        else:
            return self.new_series(
                inputs,
                shape=q_val.shape,
                dtype=dtype,
                index_value=parse_index(
                    index_val,
                    a,
                    q_val,
                    self._interpolation,
                    type(self).__name__,
                    store_data=store_index_value,
                ),
                name=a.name,
            )

    def __call__(self, a, q_input=None):
        inputs = [a]
        if q_input is not None:
            inputs.append(q_input)
        if isinstance(a, DATAFRAME_TYPE):
            return self._call_dataframe(a, inputs)
        else:
            return self._call_series(a, inputs)

    @classmethod
    def _tile_dataframe(cls, op):
        from ...tensor.merge.stack import TensorStack

        df = op.outputs[0]
        if df.ndim == 1:
            if op.axis == 0:
                ts = []
                for name in df.index_value.to_pandas():
                    a = tensor_from_series(op.input[name])
                    t = tensor_quantile(
                        a,
                        op.q,
                        interpolation=op.interpolation,
                        handle_non_numeric=not op.numeric_only,
                    )
                    ts.append(t)
                try:
                    dtype = np.result_type(*[it.dtype for it in ts])
                except TypeError:
                    dtype = np.dtype(object)
                stack_op = TensorStack(axis=0, dtype=dtype)
                tr = stack_op(ts)
                r = series_from_tensor(
                    tr, index=df.index_value.to_pandas(), name=np.asscalar(ts[0].op.q)
                )
            else:
                assert op.axis == 1
                empty_df = build_empty_df(op.input.dtypes)
                fields = empty_df._get_numeric_data().columns.tolist()
                t = tensor_from_dataframe(op.input[fields])
                tr = tensor_quantile(
                    t,
                    op.q,
                    axis=1,
                    interpolation=op.interpolation,
                    handle_non_numeric=not op.numeric_only,
                )
                r = series_from_tensor(
                    tr, index=op.input.index, name=np.asscalar(tr.op.q)
                )
        else:
            assert df.ndim == 2
            if op.axis == 0:
                d = OrderedDict()
                for name in df.dtypes.index:
                    a = tensor_from_series(op.input[name])
                    t = tensor_quantile(
                        a,
                        op.q,
                        interpolation=op.interpolation,
                        handle_non_numeric=not op.numeric_only,
                    )
                    d[name] = t
                r = create_df(d, index=op.q)
            else:
                assert op.axis == 1
                empty_df = build_empty_df(op.input.dtypes)
                fields = empty_df._get_numeric_data().columns.tolist()
                t = tensor_from_dataframe(op.input[fields])
                tr = tensor_quantile(
                    t,
                    op.q,
                    axis=1,
                    interpolation=op.interpolation,
                    handle_non_numeric=not op.numeric_only,
                )
                if not op.input.index_value.has_value():
                    raise NotImplementedError
                # TODO(xuye.qin): use index=op.input.index when we support DataFrame.index
                r = dataframe_from_tensor(
                    tr, index=op.q, columns=op.input.index_value.to_pandas()
                )

        return (yield from recursive_tile(r))

    @classmethod
    def _tile_series(cls, op):
        a = tensor_from_series(op.input)
        t = tensor_quantile(
            a,
            op.q,
            interpolation=op.interpolation,
            handle_non_numeric=not op.numeric_only,
        )
        if isinstance(op.outputs[0], TENSOR_TYPE):
            r = t
        else:
            r = series_from_tensor(t, index=op.q, name=op.outputs[0].name)
        r = yield from recursive_tile(r)
        return [r]

    @classmethod
    def tile(cls, op):
        if isinstance(op.input, DATAFRAME_TYPE):
            tiled = yield from cls._tile_dataframe(op)
        else:
            tiled = yield from cls._tile_series(op)
        return tiled


def quantile_series(series, q=0.5, interpolation="linear"):
    """
    Return value at the given quantile.

    Parameters
    ----------
    q : float or array-like, default 0.5 (50% quantile)
        0 <= q <= 1, the quantile(s) to compute.
    interpolation : {'linear', 'lower', 'higher', 'midpoint', 'nearest'}

        This optional parameter specifies the interpolation method to use,
        when the desired quantile lies between two data points `i` and `j`:

            * linear: `i + (j - i) * fraction`, where `fraction` is the
              fractional part of the index surrounded by `i` and `j`.
            * lower: `i`.
            * higher: `j`.
            * nearest: `i` or `j` whichever is nearest.
            * midpoint: (`i` + `j`) / 2.

    Returns
    -------
    float or Series
        If ``q`` is an array or a tensor, a Series will be returned where the
        index is ``q`` and the values are the quantiles, otherwise
        a float will be returned.

    See Also
    --------
    core.window.Rolling.quantile
    numpy.percentile

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 2, 3, 4])
    >>> s.quantile(.5).execute()
    2.5
    >>> s.quantile([.25, .5, .75]).execute()
    0.25    1.75
    0.50    2.50
    0.75    3.25
    dtype: float64
    """

    if isinstance(q, ENTITY_TYPE):
        q = astensor(q)
        q_input = q
    else:
        q_input = None

    op = DataFrameQuantile(q=q, interpolation=interpolation, gpu=series.op.gpu)
    return op(series, q_input=q_input)


def quantile_dataframe(df, q=0.5, axis=0, numeric_only=True, interpolation="linear"):
    """
    Return values at the given quantile over requested axis.

    Parameters
    ----------
    q : float or array-like, default 0.5 (50% quantile)
        Value between 0 <= q <= 1, the quantile(s) to compute.
    axis : {0, 1, 'index', 'columns'} (default 0)
        Equals 0 or 'index' for row-wise, 1 or 'columns' for column-wise.
    numeric_only : bool, default True
        If False, the quantile of datetime and timedelta data will be
        computed as well.
    interpolation : {'linear', 'lower', 'higher', 'midpoint', 'nearest'}
        This optional parameter specifies the interpolation method to use,
        when the desired quantile lies between two data points `i` and `j`:
        * linear: `i + (j - i) * fraction`, where `fraction` is the
          fractional part of the index surrounded by `i` and `j`.
        * lower: `i`.
        * higher: `j`.
        * nearest: `i` or `j` whichever is nearest.
        * midpoint: (`i` + `j`) / 2.

    Returns
    -------
    Series or DataFrame
        If ``q`` is an array or a tensor, a DataFrame will be returned where the
          index is ``q``, the columns are the columns of self, and the
          values are the quantiles.
        If ``q`` is a float, a Series will be returned where the
          index is the columns of self and the values are the quantiles.

    See Also
    --------
    core.window.Rolling.quantile: Rolling quantile.
    numpy.percentile: Numpy function to compute the percentile.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame(np.array([[1, 1], [2, 10], [3, 100], [4, 100]]),
    ...                   columns=['a', 'b'])
    >>> df.quantile(.1).execute()
    a    1.3
    b    3.7
    Name: 0.1, dtype: float64

    >>> df.quantile([.1, .5]).execute()
           a     b
    0.1  1.3   3.7
    0.5  2.5  55.0

    Specifying `numeric_only=False` will also compute the quantile of
    datetime and timedelta data.

    >>> df = md.DataFrame({'A': [1, 2],
    ...                    'B': [md.Timestamp('2010'),
    ...                          md.Timestamp('2011')],
    ...                    'C': [md.Timedelta('1 days'),
    ...                          md.Timedelta('2 days')]})
    >>> df.quantile(0.5, numeric_only=False).execute()
    A                    1.5
    B    2010-07-02 12:00:00
    C        1 days 12:00:00
    Name: 0.5, dtype: object
    """
    if isinstance(q, ENTITY_TYPE):
        q = astensor(q)
        q_input = q
    else:
        q_input = None
    axis = validate_axis(axis, df)

    op = DataFrameQuantile(
        q=q,
        interpolation=interpolation,
        axis=axis,
        numeric_only=numeric_only,
        gpu=df.op.gpu,
    )
    return op(df, q_input=q_input)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, AnyField
from ...core import ENTITY_TYPE
from ..array_utils import as_same_device, device
from ..core import Tensor
from ..utils import broadcast_shape
from ..datasource import tensor as astensor
from .core import TensorOperand, TensorElementWise, filter_inputs


class TensorClip(TensorOperand, TensorElementWise):
    _op_type_ = OperandDef.CLIP

    _a = KeyField("a")
    _a_min = AnyField("a_min")
    _a_max = AnyField("a_max")
    _out = KeyField("out")

    def __init__(self, a=None, a_min=None, a_max=None, out=None, **kw):
        super().__init__(_a=a, _a_min=a_min, _a_max=a_max, _out=out, **kw)

    @property
    def a(self):
        return self._a

    @property
    def a_min(self):
        return self._a_min

    @property
    def a_max(self):
        return self._a_max

    @property
    def out(self):
        return getattr(self, "_out", None)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._a = next(inputs_iter)
        if isinstance(self._a_min, ENTITY_TYPE):
            self._a_min = next(inputs_iter)
        if isinstance(self._a_max, ENTITY_TYPE):
            self._a_max = next(inputs_iter)
        if getattr(self, "_out", None) is not None:
            self._out = next(inputs_iter)

    def __call__(self, a, a_min, a_max, out=None):
        a = astensor(a)
        tensors = [a]
        sparse = a.issparse()

        if isinstance(a_min, Number):
            if a_min > 0:
                sparse = False
            a_min_dtype = np.array(a_min).dtype
        elif a_min is not None:
            a_min = astensor(a_min)
            tensors.append(a_min)
            if not a_min.issparse():
                sparse = False
            a_min_dtype = a_min.dtype
        else:
            a_min_dtype = None
        self._a_min = a_min

        if isinstance(a_max, Number):
            if a_max < 0:
                sparse = False
            a_max_dtype = np.array(a_max).dtype
        elif a_max is not None:
            a_max = astensor(a_max)
            tensors.append(a_max)
            if not a_max.issparse():
                sparse = False
            a_max_dtype = a_max.dtype
        else:
            a_max_dtype = None
        self._a_max = a_max

        if out is not None:
            if isinstance(out, Tensor):
                self._out = out
            else:
                raise TypeError(f"out should be Tensor object, got {type(out)} instead")

        dtypes = [dt for dt in [a.dtype, a_min_dtype, a_max_dtype] if dt is not None]
        dtype = np.result_type(*dtypes)
        # check broadcast
        shape = broadcast_shape(*[t.shape for t in tensors])

        setattr(self, "sparse", sparse)
        inputs = filter_inputs([a, a_min, a_max, out])
        t = self.new_tensor(inputs, shape)

        if out is None:
            setattr(self, "dtype", dtype)
            return t

        # if `out` is specified, use out's dtype and shape
        out_shape, out_dtype = out.shape, out.dtype

        if t.shape != out_shape:
            t = self.new_tensor(inputs, out_shape)
        setattr(self, "dtype", out_dtype)

        out.data = t.data
        return out

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        inputs_iter = iter(inputs)
        a = next(inputs_iter)
        a_min = (
            next(inputs_iter) if isinstance(op.a_min, type(op.outputs[0])) else op.a_min
        )
        a_max = (
            next(inputs_iter) if isinstance(op.a_max, type(op.outputs[0])) else op.a_max
        )
        out = next(inputs_iter).copy() if op.out is not None else None

        with device(device_id):
            kw = {}
            if out is not None:
                kw["out"] = out
            ctx[op.outputs[0].key] = xp.clip(a, a_min, a_max, **kw)


def clip(a, a_min, a_max, out=None):
    """
    Clip (limit) the values in a tensor.

    Given an interval, values outside the interval are clipped to
    the interval edges.  For example, if an interval of ``[0, 1]``
    is specified, values smaller than 0 become 0, and values larger
    than 1 become 1.

    Parameters
    ----------
    a : array_like
        Tensor containing elements to clip.
    a_min : scalar or array_like or `None`
        Minimum value. If `None`, clipping is not performed on lower
        interval edge. Not more than one of `a_min` and `a_max` may be
        `None`.
    a_max : scalar or array_like or `None`
        Maximum value. If `None`, clipping is not performed on upper
        interval edge. Not more than one of `a_min` and `a_max` may be
        `None`. If `a_min` or `a_max` are array_like, then the three
        arrays will be broadcasted to match their shapes.
    out : Tensor, optional
        The results will be placed in this tensor. It may be the input
        array for in-place clipping.  `out` must be of the right shape
        to hold the output.  Its type is preserved.

    Returns
    -------
    clipped_array : Tensor
        An tensor with the elements of `a`, but where values
        < `a_min` are replaced with `a_min`, and those > `a_max`
        with `a_max`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.arange(10)
    >>> mt.clip(a, 1, 8).execute()
    array([1, 1, 2, 3, 4, 5, 6, 7, 8, 8])
    >>> a.execute()
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> mt.clip(a, 3, 6, out=a).execute()
    array([3, 3, 3, 3, 4, 5, 6, 6, 6, 6])
    >>> a = mt.arange(10)
    >>> a.execute()
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    >>> mt.clip(a, [3, 4, 1, 1, 1, 4, 4, 4, 4, 4], 8).execute()
    array([3, 4, 2, 3, 4, 5, 6, 7, 8, 8])

    """
    op = TensorClip(a=a, a_min=a_min, a_max=a_max, out=out)
    return op(a, a_min, a_max, out=out)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Int32Field, StringField
from ...storage.base import StorageLevel
from ...utils import calc_nsplits, has_unknown_shape, lazy_import
from ...core.context import get_context
from ..operands import TensorOperand, TensorOperandMixin
from .core import TensorNoInput

vineyard = lazy_import("vineyard")
vy_data_utils = lazy_import("vineyard.data.utils", rename="vy_data_utils")


def resolve_vineyard_socket(ctx, op):
    if op.vineyard_socket is None:  # pragma: no cover
        storage_backend = ctx.get_storage_info(level=StorageLevel.MEMORY)
        if storage_backend.get("name", None) == "vineyard":
            return storage_backend["socket"]
        else:
            return op.vineyard_socket
    else:
        return op.vineyard_socket


class TensorFromVineyard(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_FROM_VINEYARD_META

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    # ObjectID in vineyard
    object_id = StringField("object_id")

    # a dummy attr to make sure ops have different keys
    operator_index = Int32Field("operator_index")

    def __init__(self, vineyard_socket=None, object_id=None, **kw):
        super().__init__(vineyard_socket=vineyard_socket, object_id=object_id, **kw)

    @classmethod
    def tile(cls, op):
        ctx = get_context()
        workers = ctx.get_worker_addresses()

        out_chunks = []
        for index, worker in enumerate(workers):
            chunk_op = op.copy().reset_key()
            chunk_op.expect_worker = worker
            chunk_op.operator_index = index
            out_chunk = chunk_op.new_chunk(
                [], dtype=np.dtype(object), shape=(1,), index=(index,)
            )
            out_chunks.append(out_chunk)

        new_op = op.copy().reset_key()
        return new_op.new_tensors(
            op.inputs,
            shape=(np.nan,),
            dtype=np.dtype(object),
            chunks=out_chunks,
            nsplits=((np.nan,) * len(workers),),
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        meta = client.get_meta(vineyard.ObjectID(op.object_id))
        chunks = []
        for idx in range(meta["partitions_-size"]):
            chunk_meta = meta["partitions_-%d" % idx]
            if not chunk_meta.islocal:
                continue
            dtype = vy_data_utils.normalize_dtype(
                chunk_meta["value_type_"], chunk_meta.get("value_type_meta_", None)
            )
            shape = tuple(json.loads(chunk_meta["shape_"]))
            chunk_index = tuple(json.loads(chunk_meta["partition_index_"]))
            # chunk: (chunk_id, worker_address, dtype, shape, index)
            chunks.append(
                (repr(chunk_meta.id), ctx.worker_address, dtype, shape, chunk_index)
            )

        holder = np.empty((1,), dtype=object)
        holder[0] = chunks
        ctx[op.outputs[0].key] = np.asarray(holder)


class TensorFromVineyardChunk(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.TENSOR_FROM_VINEYARD_CHUNK

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    # ObjectID of chunk in vineyard
    object_id = StringField("object_id")

    def __init__(self, vineyard_socket=None, object_id=None, **kw):
        super().__init__(vineyard_socket=vineyard_socket, object_id=object_id, **kw)

    def __call__(self, meta):
        return self.new_tensor([meta], shape=(np.nan,))

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield

        ctx = get_context()

        in_chunk_keys = [chunk.key for chunk in op.inputs[0].chunks]
        out_chunks = []
        chunk_map = dict()
        dtype = None
        for chunk, infos in zip(
            op.inputs[0].chunks, ctx.get_chunks_result(in_chunk_keys)
        ):
            for info in infos[0]:  # n.b. 1-element ndarray
                chunk_op = op.copy().reset_key()
                chunk_op.object_id = info[0]
                chunk_op.expect_worker = info[1]
                dtype = info[2]
                shape = info[3]
                chunk_index = info[4]
                chunk_map[chunk_index] = info[3]
                out_chunk = chunk_op.new_chunk(
                    [chunk], shape=shape, dtype=dtype, index=chunk_index
                )
                out_chunks.append(out_chunk)

        nsplits = calc_nsplits(chunk_map)
        shape = [np.sum(nsplit) for nsplit in nsplits]
        new_op = op.copy().reset_key()
        return new_op.new_tensors(
            op.inputs, shape=shape, dtype=dtype, chunks=out_chunks, nsplits=nsplits
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        client = vineyard.connect(socket)
        ctx[op.outputs[0].key] = client.get(vineyard.ObjectID(op.object_id))


def fromvineyard(tensor, vineyard_socket=None):
    if vineyard is not None and isinstance(tensor, vineyard.Object):  # pragma: no cover
        if "vineyard::GlobalTensor" not in tensor.typename:
            raise TypeError(
                "The input tensor %r is not a vineyard' GlobalTensor" % tensor
            )
        object_id = tensor.id
    else:
        object_id = tensor
    if vineyard is not None and isinstance(object_id, vineyard.ObjectID):
        object_id = repr(object_id)
    metaop = TensorFromVineyard(
        vineyard_socket=vineyard_socket,
        object_id=object_id,
        dtype=np.dtype("byte"),
        gpu=None,
    )
    meta = metaop(shape=(np.nan,), chunk_size=(np.nan,))
    op = TensorFromVineyardChunk(
        vineyard_socket=vineyard_socket, object_id=object_id, gpu=None
    )
    return op(meta)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import contextlib
import sys

from ... import oscar as mo
from ...lib.aio import alru_cache
from ..subtask import SubtaskResult, SubtaskStatus
from ..task import TaskAPI


@alru_cache
async def _get_task_api(actor: mo.Actor):
    return await TaskAPI.create(getattr(actor, "_session_id"), actor.address)


@contextlib.asynccontextmanager
async def redirect_subtask_errors(actor: mo.Actor, subtasks):
    try:
        yield
    except:  # noqa: E722  # pylint: disable=bare-except
        _, error, traceback = sys.exc_info()
        status = (
            SubtaskStatus.cancelled
            if isinstance(error, asyncio.CancelledError)
            else SubtaskStatus.errored
        )
        task_api = await _get_task_api(actor)
        coros = []
        for subtask in subtasks:
            if subtask is None:  # pragma: no cover
                continue
            coros.append(
                task_api.set_subtask_result(
                    SubtaskResult(
                        subtask_id=subtask.subtask_id,
                        session_id=subtask.session_id,
                        task_id=subtask.task_id,
                        stage_id=subtask.stage_id,
                        progress=1.0,
                        status=status,
                        error=error,
                        traceback=traceback,
                    )
                )
            )
        tasks = [asyncio.ensure_future(coro) for coro in coros]
        await asyncio.wait(tasks)
        raise
